'''
Created on 6 sept. 2017

@author: Fab
'''
import sqlite3
from time import *
from lmtanalysis.Chronometer import Chronometer
from lmtanalysis.Animal import *
from lmtanalysis.Detection import *
from lmtanalysis.Measure import *
import numpy as np
from lmtanalysis.Event import *
from lmtanalysis.Measure import *
#from affine import Affine
import matplotlib.pyplot as plt
import matplotlib.lines as mlines

def flush( connection ):
    ''' flush event in database '''
    deleteEventTimeLineInBase(connection, "Center Zone" )
    deleteEventTimeLineInBase(connection, "Periphery Zone" )
    

def reBuildEvent( connection, file, tmin=None, tmax=None, pool = None ): 
    
    ''' use the pool provided or create it'''
    if ( pool == None ):
        pool = AnimalPool( )
        pool.loadAnimals( connection )
        pool.loadDetection( start = tmin, end = tmax )
    '''
    Event Center
    - the animal is in the center zone of the cage
    center zone: xa=149, xb=363, ya=318, yb=98
    
    Event Periphery
    - the animal is at the periphery of the cage (opposite event from Center)
    '''
    
    for animal in pool.animalDictionnary.keys():
        # print(pool.animalDictionnary[animal])
        
        eventNameCenter = "Center Zone"
        # print ( "A is in the center zone")
        # print ( eventNameCenter )
        
        eventNamePeriphery = "Periphery Zone"
        # print ( "A is in the periphery zone")
        # print ( eventNamePeriphery )
                
        centerZoneTimeLine = EventTimeLine( None, eventNameCenter , animal , None , None , None , loadEvent=False )
        peripheryZoneTimeLine = EventTimeLine( None, eventNamePeriphery , animal , None , None , None , loadEvent=False )
                
        resultCenter={}
        resultPeriphery={}
        
        animalA = pool.animalDictionnary[animal]
        #print ( animalA )
        dicA = animalA.detectionDictionnary
            
        for t in dicA.keys():

            if (dicA[t].isInZone(xa=168, xb=343, ya=296, yb=120) == True):
                resultCenter[t] = True
                
            else:
                resultPeriphery[t] = True
                
        centerZoneTimeLine.reBuildWithDictionnary( resultCenter )
        centerZoneTimeLine.endRebuildEventTimeLine(connection)
    
        peripheryZoneTimeLine.reBuildWithDictionnary( resultPeriphery )
        peripheryZoneTimeLine.endRebuildEventTimeLine(connection)
        
    # log process
    from lmtanalysis.TaskLogger import TaskLogger
    t = TaskLogger( connection )
    t.addLog( "Build Event Center Zone" , tmin=tmin, tmax=tmax )
    t.addLog( "Build Event Periphery Zone" , tmin=tmin, tmax=tmax )

    # print( "Rebuild event finished." )
        
        
        
        
        
        
        
    


'''
Created on 6 sept. 2017

@author: Fab
'''
import sqlite3
from time import *
from lmtanalysis.Chronometer import Chronometer
from lmtanalysis.Animal import *
from lmtanalysis.Detection import *
from lmtanalysis.Measure import *
import matplotlib.pyplot as plt
import numpy as np
from lmtanalysis.Event import *
from lmtanalysis.EventTimeLineCache import EventTimeLineCached


class Train2():
    
    def __init__(self, idA, idB ):
        self.idA = idA
        self.idB = idB


def flush( connection ):
    ''' flush event in database '''
    deleteEventTimeLineInBase(connection, "Train3" )

def reBuildEvent( connection, file, tmin=None, tmax=None , pool = None ): 

    ''' use pool cache if available '''
    if ( pool == None ):
        pool = AnimalPool( )
        pool.loadAnimals( connection )
        if ( len ( pool.getAnimalList( ) ) < 3 ):
            # print("Train 3 cannot be computed on an experiment with less than 3 animals")
            return
        pool.loadDetection( start = tmin, end = tmax )    
    
    '''
    three animals are following each others with nose-to-anogenital contacts
    animals are moving
    
    this is a combination of Train2 events (train2 events must be calculated before this event)
    '''
    
    #deleteEventTimeLineInBase(connection, "Train3" )

    ''' build a list of train 2 for each time point '''
    
    time = {}
    train3 = {}
    
    for animal in range( 1,5 ):
        for idAnimalB in range( 1 , 5 ):
            if ( animal == idAnimalB ):
                continue
            train2TimeLine = EventTimeLineCached( connection, file, "Train2", animal, idAnimalB, minFrame=tmin, maxFrame=tmax )
            for t in train2TimeLine.getDictionnary():
                train = Train2( animal, idAnimalB )
                
                if ( not t in time ):
                    time[t] =[]
                    
                #print ( t , ":" , train.idA , " -> ", train.idB , "*" )
                time[t].append( train )
    

    for t in time:
        trainList = time[t]
        
        
        for trainSource in trainList:
            
            for trainTarget in trainList:
                
                if ( trainSource == trainTarget ):
                    continue
                
                isValid = ""
                ''' test chain link between train2 events '''
               
                if trainSource.idB == trainTarget.idA:
                    
                    id1 = trainSource.idA
                    id2 = trainSource.idB
                    id3 = trainTarget.idB

                    if not (id1, id2, id3) in train3:    
                        train3[id1,id2,id3] = {} 
                    
                    train3[id1,id2,id3][t]=True
                    
                    isValid = ": validated train 3"
                #print ( t , ":" , trainSource.idA , " -> ", trainSource.idB, "--->", trainTarget.idA , " -> ", trainTarget.idB , isValid )


    ''' save data '''
            
    for animal in range( 1 , 5 ):
    
        for idAnimalB in range( 1 , 5 ):
            
            for idAnimalC in range( 1 , 5 ):

                if (animal, idAnimalB, idAnimalC) in train3:
                    
                    trainTimeLine = EventTimeLine( None, "Train3" , animal , idAnimalB, idAnimalC, loadEvent=False )
                    
                    trainTimeLine.reBuildWithDictionnary( train3[animal,idAnimalB,idAnimalC] )
                    #trainTimeLine.removeEventsBelowLength( 5 )            
                    trainTimeLine.endRebuildEventTimeLine(connection)
    
        
    # log process
    from lmtanalysis.TaskLogger import TaskLogger
    t = TaskLogger( connection )
    t.addLog( "Build Event Train3" , tmin=tmin, tmax=tmax )

                    
    # print( "Rebuild event finished." )
    
    return
   
    
    


'''
Created on 6 sept. 2017

@author: Fab
'''
import sqlite3
from time import *
from lmtanalysis.Chronometer import Chronometer
from lmtanalysis.Animal import *
from lmtanalysis.Detection import *
from lmtanalysis.Measure import *
import numpy as np
from lmtanalysis.Event import *
from lmtanalysis.Measure import *
#from affine import Affine
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
from lmtanalysis.EventTimeLineCache import EventTimeLineCached

def flush( connection ):
    ''' flush event in database '''
    deleteEventTimeLineInBase(connection, "Rear isolated" )
    deleteEventTimeLineInBase(connection, "Rear in contact" )


def reBuildEvent( connection, file, tmin=None, tmax=None, pool = None ): 
    
    ''' use the pool provided or create it'''
    if ( pool == None ):
        pool = AnimalPool( )
        pool.loadAnimals( connection )
        pool.loadDetection( start = tmin, end = tmax )
    
    '''
    Event Rear5:
    - the animal is rearing
    - distinction between rearing in contact with one or several animals and rearing isolated from the others
    '''
        
    contact = {}
    
    for animal in pool.animalDictionnary.keys():
        # print(pool.animalDictionnary[animal])
        
        contact[animal] = EventTimeLineCached( connection, file, "Contact", animal, minFrame=tmin, maxFrame=tmax )
        contactDico = contact[animal].getDictionnary()
        
        eventName1 = "Rear isolated"
        eventName2 = "Rear in contact"
        # print ( "A rears")
                
        rearSocialTimeLine = EventTimeLine( None, eventName2 , animal , None , None , None , loadEvent=False )
        rearIsolatedTimeLine = EventTimeLine( None, eventName1 , animal , None , None , None , loadEvent=False )
                
        resultSocial={}
        resultIsolated={}
        
        animalA = pool.animalDictionnary[animal]
        #print ( animalA )
        dicA = animalA.detectionDictionnary
            
        for t in dicA.keys():
            
            slope = dicA[t].getBodySlope()
            if ( slope == None):
                continue
            
            if ( abs( slope ) < BODY_SLOPE_THRESHOLD ):
                continue;
                
            if (t in contactDico.keys()):
                #print("social")
                resultSocial[t] = True
            
            else:
                #print("isolated")
                resultIsolated[t] = True
    
        rearSocialTimeLine.reBuildWithDictionnary( resultSocial )
        rearIsolatedTimeLine.reBuildWithDictionnary( resultIsolated )
        
        rearSocialTimeLine.endRebuildEventTimeLine( connection )
        rearIsolatedTimeLine.endRebuildEventTimeLine( connection )
        
        
    # log process
    from lmtanalysis.TaskLogger import TaskLogger
    t = TaskLogger( connection )
    t.addLog( "Build Event Rear5" , tmin=tmin, tmax=tmax )

                
    # print( "Rebuild event finished." )
        
        
        
        
        
        
        
    


"""
Created on 20 August 2019 + update in 2020 by D. Huzard + updated in 2023 by Paul Carrascosa

@author: D. Huzard, P. Carrascosa
"""

"""
This code will extract all the behaviors into timebins and export it in .csv files.
The Goal is to then use an ipywidget tool, based on dataframes to compute Graphs and mixed model statistics.
"""

import sys
sys.path.insert(1, "../")

import sqlite3
import warnings
warnings.simplefilter(action='ignore')
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import datetime
import os
from tqdm import tqdm
from lmtanalysis.Animal import AnimalPool
from lmtanalysis.Event import EventTimeLine, plotMultipleTimeLine
from lmtanalysis.FileUtil import getFilesToProcess
from lmtanalysis.Measure import *
from IPython.display import display
import ipywidgets as widgets

def computeBehaviorsData(behavior, show=False):
    """ Computes the details (Mean, Nb, Std, CI95...) of behaviors."""

    name = behavior.eventName
    # nameAndIds = behavior.eventNameWithId
    idA = behavior.idA
    idB = behavior.idB
    idC = behavior.idC
    idD = behavior.idD
    totalLength = behavior.getTotalLength()
    meanLength = behavior.getMeanEventLength()
    numberOfEvents = behavior.getNumberOfEvent()
    stdLength = behavior.getStandardDeviationEventLength()

    """    if behavior.getMedianEventLength() is not None:
        medianLength = behavior.getMedianEventLength()
    else:
        medianLength = None"""

    medianLength = None

    CI95 = [None]*2
    if meanLength is not None:
        CI = 1.96 * behavior.getStandardDeviationEventLength() / np.math.sqrt(behavior.getNumberOfEvent())
        CI95[0] = meanLength - CI
        CI95[1] = meanLength + CI
    else:
        meanLength = None

    returnedBehaviors = {
        "name": name,
        "idA": idA,
        "idB": idB,
        "idC": idC,
        "idD": idD,
        "totalLength": totalLength,
        "meanLength": meanLength,
        "medianLength": medianLength,
        "numberOfEvents": numberOfEvents,
        "stdLength": stdLength,
        "CI95_low": CI95[0],
        "CI95_up": CI95[1]
    }

    # print(f"{nameAndIds}")
    # print(f"   -TotalLength: {totalLength} frames.")
    # print(f"   -MeanEventLength: {meanLength}")
    # print(f"   -MedianEventLength: {medianLength}")
    # print(f"   -NumberOfEvent: {numberOfEvents}")
    # print(f"   -StandardDeviationEventLength: {stdLength}")
    # print(f"   -The confidence interval (95%) is [{CI95[0]},{CI95[1]}].")

    # if show:
        # beh.plotEventDurationDistributionHist(nbBin=30, title="Timebin #"+str(bin) + " / " + beh.eventNameWithId)
        # beh.plotEventDurationDistributionBar(title="Timebin #"+str(bin) + " / " + beh.eventNameWithId)

    # TODO: Show the histogram of events, with a 'show' parameter, with one subplot per behavior

    return returnedBehaviors

def Export():
    files = getFilesToProcess()
    # if len(files) == 0:
    #     print("NO FILE TO PROCESS !!!!!")
    # if len(files) >= 1:
    filenames = [os.path.basename(files[x]) for x in range(0, len(files))]
    # print(f"{files} => {filenames}")

    ### DEFINE CONSTANTS ###
    start = {}
    stop = {}

    timeBinsDuration = int(input("Enter the TIMEBIN for ALL the files (1min =  1800 frames / 1h = 108000 frames): "))

    # fileGlobal = input("Enter the filename for the .csv WITH ALL DATA INSIDE: ")

    useNights = input("Do you want to use the Nights from the .sqlite files to computes the data ? ('Yes'/'No'): ")

    """
    useSameStartStop = input("Do you want to use the same start & stop frames for all DBs ? ('Yes'/'No'): ")

    if useSameStartStop.lower() == "yes":
        startAll = int(input(f"Enter the Global STARTING frame:"))
        stopAll = int(input(f"Enter the Global ENDING frame:"))
        nbTimebins = int((stopAll - startAll) / timeBinsDuration)
        print(f"There are {nbTimebins} timebins of {timeBinsDuration} frames (= {timeBinsDuration / 30 / 60} "
              f"minutes) between frames: Start={startAll} and Stop={stopAll}.")
        for filename in filenames:
            start[filename] = startAll
            stop[filename] = stopAll

    if useSameStartStop.lower() == "no":
        for filename in filenames:
            start[filename] = int(input(f"Enter the STARTING frame for {filename}:"))
            stop[filename] = int(input(f"Enter the ENDING frame for {filename}:"))
            nbTimebins = int((stop[filename] - start[filename]) / timeBinsDuration)
            print(f"There are {nbTimebins} timebins of {timeBinsDuration} frames (= {timeBinsDuration/30/60} minutes) "
                  f"between frames: Start={start[filename]} and Stop={stop[filename]}.")
    """

    #Create a Global Dataframe with all data:
    dfGlobal = pd.DataFrame()

    count = 0
    for file in files:
        # print(f"The current Count is : {count}")
        # print(f"file: {file}")
        # print(f"File path: {file.title()}")
        fileName = filenames[count]
        print(f"File name: {fileName}")

        #Check that 'filename' and the current file are the same
        #TODO CAN BE IMPROVED by extracting from the current file

        # if fileName in file:
        #     print("THE FILE NAME MATCHES! !")
        # else:
        #     print("!!! ERROR: FILE NAME DO NOT MATCH!!")

        connection = sqlite3.connect(file) # connect to database
        animalPool = AnimalPool() # create an animalPool, which basically contains your animals
        animalPool.loadAnimals(connection) # load infos about the animals

        animalNumber = animalPool.getNbAnimals()
        # print(f"There are {animalNumber} animals,")

        if useNights.lower() == "yes" or useNights.lower() == "y":
            # Le problème vient d'ici (fonction getNightStartStop) ? Est-ce que ça ne prendrait pas juste la première
            # nuit ? Faire une boucle if qui dit que si night_phase > 1 alors on prendra, à chaque night_count,
            # les données des frames suivantes comme getNightStartStop ne prend que le startframe et le endframe de
            # l'évènement night ?
            NightFrames = animalPool.getNightStartStop()
            # print("The night events are:")
            # print(NightFrames)
            #print(NightFrames[0])

            #TODO find number of night and create bins for each

            nbTimebin = []

            startFrame = NightFrames[0][0]
            stopFrame = NightFrames[0][1]

            nbTimebin.append(int((stopFrame - startFrame) / timeBinsDuration)+1)
            # print("There are "+str(nbTimebin)+" timebins.")
            #
            # print(startFrame)
            # print(stopFrame)

            # for night in NightFrames:
            #     print(night)

            # for night in NightFrames:
            #     print(night)
            #     print(night[0])
            #     print(night[1])

            nbTimebins = int((stopFrame - startFrame) / timeBinsDuration)+1

            # print(f"There are {nbTimebins} timebins of {timeBinsDuration} frames (= {timeBinsDuration / 30 / 60} "
            #       f"minutes) during the night (between frames {startFrame} and {stopFrame}.)")

            for filename in filenames:
                start[filename] = startFrame
                stop[filename] = stopFrame

        if useNights.lower() == "no" or useNights.lower() == "no":
            print("AAAAaaaahhh")

        #TODO Flush ALL the events !!

        #Different events Depending on the number of animals in the Database:
        if animalNumber >= 1:
            behavioralEventsForOneAnimal = ["Move", "Move isolated", "Rearing", "Rear isolated",
                                            "Stop isolated", "WallJump", "SAP", "Huddling", "WaterPoint", "Distance"]
            # print("The behaviors extracted are:\n", behavioralEventsForOneAnimal)
        if animalNumber >= 2:
            behavioralEventsForTwoAnimals = ["Contact", "Oral-oral Contact", "Oral-genital Contact",
                                             "Side by side Contact",
                                             "Side by side Contact, opposite way", "Social approach", "Social escape",
                                             "Approach contact", "Approach rear", "Break contact", "Get away",
                                             "FollowZone Isolated",
                                             "Train2", "Group2", "Move in contact", "Rear in contact"]
            # print("and:\n", behavioralEventsForTwoAnimals)
        if animalNumber >= 3:
            behavioralEventsForThreeAnimals = ["Group3", "Group 3 break", "Group 3 make"]
            # print("and:\n", behavioralEventsForThreeAnimals)
        if animalNumber >= 4:
            behavioralEventsForFourAnimals = ["Group4", "Group 4 break", "Group 4 make", "Nest3", "Nest4"]
            # print("and:", behavioralEventsForFourAnimals)

        # Export the Infos (Mean, Std, ...) of the Behaviors for each timebin:
        dicoOfBehInfos = {
            "Filename": None,
            "Date": None,
            "Cage": None,
            "Injection": None,
            "Night-Phase": None,
            "Bin": None,
            "start_frame": None,
            "stop_frame": None,
            "name": None,
            "idA": None,
            "idB": None,
            "idC": None,
            "idD": None,
            "RFidA": None,
            "RFidB": None,
            "RFidC": None,
            "RFidD": None,
            "GenoA": None,
            "GenoB": None,
            "GenoC": None,
            "GenoD": None,
            "totalLength": None,
            "meanLength": None,
            "medianLength": None,
            "numberOfEvents": None,
            "stdLength": None,
            "CI95_low": None,
            "CI95_up": None
        }

        dfOfBehInfos = pd.DataFrame()  # Dataframe with the details of the previous dict

        allBehaviorsInfo = {}

        #TODO one for loop on Nights and One for loop for the times bins inside the night
        night_count = 1
        # print(NightFrames)
        for night in NightFrames:
            # print("The night is ", night)
            # print("The night_ count is ", night_count)
            bin = 1
            bin_progress = tqdm(range(1, nbTimebins + 1), desc=f"Night {night_count} - Bins", position=0, leave=True)
            for z in bin_progress:
                # print("Z is ", z)
                startBin = (start[fileName] + (z-1) * timeBinsDuration) + (night_count-1) * (108000*24)
                stopBin = startBin + timeBinsDuration
                bin_progress.set_postfix(Bin=f"{bin}/{nbTimebins}")
        # for night in NightFrames:
        #     # print("The night is ", night)
        #     # print("The night_ count is ", night_count)
        #     bin = 1
        #     for z in range(night[0], night[1], timeBinsDuration):
        #         # print("Z is ", z)
        #         startBin = (start[fileName] + (bin-1) * timeBinsDuration) + (night_count-1) * (108000*24)
        #         stopBin = startBin + timeBinsDuration
                # print(bin)
            # for bin in range(nbTimebins):

                # print(f"************* Loading data for bin #{bin} *************")
                # now = datetime.datetime.now()
                # print("Current date and time : ")
                # print(now.strftime("%Y-%m-%d %H:%M:%S"))
                # print(f"*** Start frame = {startBin} // Stop frame = {stopBin} ***")
                animalPool.loadDetection(start=startBin, end=stopBin)  # load the detection for the different bins

                #Update the dictionary with Start; Stop; and bin info:
                dicoOfBehInfos["start_frame"] = startBin
                dicoOfBehInfos["stop_frame"] = stopBin
                dicoOfBehInfos["Bin"] = bin
                dicoOfBehInfos["Night-Phase"] = night_count

                #Update the dictionary with Filename; Date; Cage; and Session info
                # Example file name "191104_Magel2_Cage4_LMT3_4mo - Copy.sqlite"
                dicoOfBehInfos["Filename"] = fileName[:-7] #Remove the " - Copy.sqlite" part
                date, Xp, cage, Injection = fileName[:-7].split("_")  # Splits the name at '_'
                # print(f"{date} , {Xp}, {cage}, {Injection}")
                dicoOfBehInfos["Date"] = date
                dicoOfBehInfos["Cage"] = cage
                dicoOfBehInfos["Injection"] = Injection

                # print("****** General parameters of the animals ****** ")
                # for animal in animalPool.getAnimalList():
                    # print("**")
                    # print(f"Animal RFID: {animal.RFID} / Animal Id: {animal.baseId} / Animal name: {animal.name}")
                    # print(f"Animal genotype: {animal.genotype}")
                    # nbOfDetectionFrames = len(animal.detectionDictionnary.keys()) timeInSecond =
                    # nbOfDetectionFrames / 30  # 30 fps print("Detection time: ", timeInSecond, "seconds.") print(
                    # "Distance traveled in arena (cm): ", animal.getDistance(tmin=start, tmax=stop))  # distance
                    # traveled

                #### For 1+ ANIMAL ####
                if animalNumber >= 1:
                    for behavior in behavioralEventsForOneAnimal:
                        # print("**** ", behavior, " ****")
                        # behavioralList1 = []

                        for a in animalPool.getAnimalDictionnary():
                            if behavior == "Distance":
                                # print("DISTANCE COMPUTE !!!!!!!!!!")
                                dist_temp = animalPool.getAnimalWithId(a).getDistance(startBin, stopBin)
                                # print(dist_temp)
                                dicoOfBehInfos.update({"name": "Distance", "totalLength": dist_temp,
                                                       "RFidA": animalPool.getAnimalWithId(a).RFID,
                                                       "GenoA": animalPool.getAnimalWithId(a).genotype,
                                                       "idA": a})
                            else:

                                eventTimeLine = EventTimeLine(connection, behavior, idA=a, minFrame=startBin,
                                                              maxFrame=stopBin)

                                behavioralDataOne = computeBehaviorsData(eventTimeLine) #Computes the Stats.
                                dicoOfBehInfos.update(behavioralDataOne) #ADD THE CALCULATIONS FROM THE EVENTIMELINE
                                dicoOfBehInfos.update({"RFidA": animalPool.getAnimalWithId(a).RFID,
                                                       "GenoA": animalPool.getAnimalWithId(a).genotype})

                            # Creates a dataFrame with the behavioral Infos from the dictionary:
                            index = [0]
                            dfOfBehInfosTemp = pd.DataFrame(dicoOfBehInfos, index=index)
                            dfOfBehInfos = dfOfBehInfos.append(dfOfBehInfosTemp, ignore_index=True)

                        # TODO for each Timebin, Plot the timeline of the behaviors ?
                        # plotMultipleTimeLine(behavioralList, title=behavior+" / timebin #"+str(bin), minValue=start)

                #### FOR 2+ ANIMALS ####
                if animalNumber >= 2:
                    for behavior in behavioralEventsForTwoAnimals:
                        # print("**** ", behavior, " ****")
                        behavioralList2 = []

                        for a in animalPool.getAnimalDictionnary():
                            if behavior == "Move in contact" or behavior == "Rear in contact":  # Just idA in those behaviors
                                eventTimeLine = EventTimeLine(connection, behavior, idA=a, minFrame=startBin, maxFrame=stopBin)
                                continue
                            for b in animalPool.getAnimalDictionnary():
                                if a == b:
                                    continue
                                eventTimeLine = EventTimeLine(connection, behavior, idA=a, idB=b,
                                                              minFrame=startBin, maxFrame=stopBin)

                                behavioralDataTwo = computeBehaviorsData(eventTimeLine)  #Computes the Stats.
                                dicoOfBehInfos.update(behavioralDataTwo)  # ADD THE CALCULATIONS FROM THE EVENTIMELINE
                                dicoOfBehInfos.update({"RFidA": animalPool.getAnimalWithId(a).RFID,
                                                       "RFidB": animalPool.getAnimalWithId(b).RFID,
                                                       "GenoA": animalPool.getAnimalWithId(a).genotype,
                                                       "GenoB": animalPool.getAnimalWithId(b).genotype})

                                # Creates a dataFrame with the behavioral Infos from the dictionary:
                                index = [0]
                                dfOfBehInfosTemp = pd.DataFrame(dicoOfBehInfos, index=index)
                                dfOfBehInfos = dfOfBehInfos.append(dfOfBehInfosTemp, ignore_index=True)

                #### FOR 3+ ANIMALS ####
                if animalNumber >= 3:
                    for behavior in behavioralEventsForThreeAnimals:
                        # print("**** ", behavior, " ****")

                        for a in animalPool.getAnimalDictionnary():
                            # There is ONLY ONE animal making or braking a group of 3 or 4 mice
                            if behavior == "Group 3 make" or behavior == "Group 3 break":
                                eventTimeLine = EventTimeLine(connection, behavior, idA=a, minFrame=startBin, maxFrame=stopBin)
                                continue
                            for b in animalPool.getAnimalDictionnary():
                                if a == b:
                                    continue
                                for c in animalPool.getAnimalDictionnary():
                                    if a == c or b == c:
                                        continue
                                    eventTimeLine = EventTimeLine(connection, behavior, idA=a, idB=b, idC=c,
                                                                  minFrame=startBin, maxFrame=stopBin)

                                    behavioralDataThree = computeBehaviorsData(eventTimeLine)  # Computes the Stats.
                                    dicoOfBehInfos.update(behavioralDataThree)  # ADD THE CALCULATIONS FROM THE EVENTIMELINE
                                    dicoOfBehInfos.update({"RFidA": animalPool.getAnimalWithId(a).RFID,
                                                           "RFidB": animalPool.getAnimalWithId(b).RFID,
                                                           "RFidC": animalPool.getAnimalWithId(c).RFID,
                                                           "GenoA": animalPool.getAnimalWithId(a).genotype,
                                                           "GenoB": animalPool.getAnimalWithId(b).genotype,
                                                           "GenoC": animalPool.getAnimalWithId(c).genotype})

                                    # Creates a dataFrame with the behavioral Infos from the dictionary:
                                    index = [0]
                                    dfOfBehInfosTemp = pd.DataFrame(dicoOfBehInfos, index=index)
                                    dfOfBehInfos = dfOfBehInfos.append(dfOfBehInfosTemp, ignore_index=True)

                #### FOR 4 ANIMALS ####
                if animalNumber >= 4:
                    for behavior in behavioralEventsForFourAnimals:
                        # print("**** ", behavior, " ****")

                        for a in animalPool.getAnimalDictionnary():
                            # There is ONLY ONE animal making or braking a group of 3 or 4 mice
                            if behavior == "Group 4 make" or behavior == "Group 4 break":
                                eventTimeLine = EventTimeLine(connection, behavior, idA=a,
                                                              minFrame=startBin, maxFrame=stopBin)
                                continue
                            for b in animalPool.getAnimalDictionnary():
                                if a == b:
                                    continue
                                for c in animalPool.getAnimalDictionnary():
                                    if a == c or b == c:
                                        continue
                                    # There is ONLY THREE animals making a Nest3
                                    if behavior == "Nest3":
                                        eventTimeLine = EventTimeLine(connection, behavior, idA=a, idB=b, idC=c,
                                                                      minFrame=startBin, maxFrame=stopBin)
                                        continue
                                    for d in animalPool.getAnimalDictionnary():
                                        if a == d or b == d or c == d:
                                            continue
                                        eventTimeLine = EventTimeLine(connection, behavior, idA=a, idB=b, idC=c, idD=d,
                                                                      minFrame=startBin, maxFrame=stopBin)

                                        behavioralDataFour = computeBehaviorsData(eventTimeLine)
                                        dicoOfBehInfos.update(behavioralDataFour)
                                        dicoOfBehInfos.update({"RFidA": animalPool.getAnimalWithId(a).RFID,
                                                               "RFidB": animalPool.getAnimalWithId(b).RFID,
                                                               "RFidC": animalPool.getAnimalWithId(c).RFID,
                                                               "RFidD": animalPool.getAnimalWithId(d).RFID,
                                                               "GenoA": animalPool.getAnimalWithId(a).genotype,
                                                               "GenoB": animalPool.getAnimalWithId(b).genotype,
                                                               "GenoC": animalPool.getAnimalWithId(c).genotype,
                                                               "GenoD": animalPool.getAnimalWithId(d).genotype})

                                        # Creates a dataFrame with the behavioral Infos from the dictionary:
                                        index = [0]
                                        dfOfBehInfosTemp = pd.DataFrame(dicoOfBehInfos, index=index)
                                        dfOfBehInfos = dfOfBehInfos.append(dfOfBehInfosTemp, ignore_index=True)

                bin += 1
            night_count += 1
        count += 1 # TODO REPLACE THIS COUNT BY A FOR LOOP ON FILES

        dfGlobal = dfGlobal.append(dfOfBehInfos)  # Add the data to the global Dataframe

        dfOfBehInfos.to_csv(f"{fileName}.csv") # Export the current dataframe into a .csv
        print(f"{fileName}.csv File Created !")
        # print("##################################################################################")
        # print("##################################################################################")
        print("######################## Close Connection with Database ##########################")
    connection.close()

    # dfGlobal.to_csv(f"{fileGlobal}.csv")  # Export the Global dataframe into a .csv
    # print(f"{fileGlobal}.csv File Created !")

    # Say it's done !
    print("!!! End of analysis !!!")

# if __name__ == '__main__':
#     Export()


'''
Created on 6 sept. 2017

@author: Fab
'''
import sqlite3
from time import *
from lmtanalysis.Chronometer import Chronometer
from lmtanalysis.Animal import *
from lmtanalysis.Detection import *
from lmtanalysis.Measure import *
import matplotlib.pyplot as plt
import numpy as np
from lmtanalysis.Event import *
from lmtanalysis.Measure import *
from lmtanalysis.EventTimeLineCache import EventTimeLineCached

def flush( connection ):
    ''' flush event in database '''
    deleteEventTimeLineInBase(connection, "Stop in contact" )
    deleteEventTimeLineInBase(connection, "Stop isolated" )



def reBuildEvent( connection, file, tmin=None, tmax=None, pool = None ):
    
    
    ''' 
    Animal A is stopped (built-in event):
    Stop in contact: animal A is stopped and in contact with another animal.
    Stop isolated: animal A is stopped and not in contact with any other animal.
    ''' 
    
    pool = AnimalPool( )
    pool.loadAnimals( connection )

    if len(pool.animalDictionnary.keys()) == 1:
        print('Only one animal in database.')
        # if the animal has been tested alone, only the stop isolated event will be computed.'''
        stopSourceTimeLine = {}

        animal = 1
        # Load source stop timeLine and revert it to get the move timeline
        # If the animal is not detected, this will result in a move. To avoid this we mask with the detection.
        
        stopSourceTimeLine[animal] = EventTimeLine(connection, "Stop", animal, minFrame=tmin, maxFrame=tmax, inverseEvent=False)
        detectionTimeLine = EventTimeLine(connection, "Detection", animal, minFrame=tmin, maxFrame=tmax)
        stopSourceTimeLine[animal].keepOnlyEventCommonWithTimeLine(detectionTimeLine)

        # initialisation of a dic for isolated stop
        stopIsolatedResult = {}

        # loop over eventlist
        for stopEvent in stopSourceTimeLine[animal].eventList:
            for t in range(stopEvent.startFrame, stopEvent.endFrame + 1):
                stopIsolatedResult[t] = True

        # save stop isolated at the end of the complete process for animalA
        stopIsolatedResultTimeLine = EventTimeLine(None, "Stop isolated", idA=animal, idB=None, idC=None, idD=None, loadEvent=False)
        stopIsolatedResultTimeLine.reBuildWithDictionnary(stopIsolatedResult)
        stopIsolatedResultTimeLine.endRebuildEventTimeLine(connection)


    else:
        # print('More than one animal in database.')
        isInContactSourceDictionnary = {}
        stopSourceTimeLine = {}
        stopIsolatedTimeLine = {}

        for animal in pool.animalDictionnary.keys():
            # Load source stop timeLine
            # If the animal is not detected, this will result in a stop. To avoid this we mask with the detection.
            
            stopSourceTimeLine[animal] = EventTimeLine( connection, "Stop", animal, minFrame=tmin, maxFrame=tmax, inverseEvent=False )
            detectionTimeLine = EventTimeLine( connection, "Detection", animal, minFrame=tmin, maxFrame=tmax )
            stopSourceTimeLine[animal].keepOnlyEventCommonWithTimeLine( detectionTimeLine )
            
            #by default let's say that all stops are isolated; next we will extract frames which are in contact from this time line.
            stopIsolatedTimeLine[animal] = stopSourceTimeLine[animal]
            
            # load contact dictionary with another animal
            for animalB in pool.animalDictionnary.keys():
                if animal == animalB:
                    # print('Same identity')
                    continue
                else:
                    isInContactSourceDictionnary[(animal, animalB)] = EventTimeLineCached( connection, file, "Contact", animal, animalB, minFrame=tmin, maxFrame=tmax ).getDictionnary()

        stopIsolatedDic = {}
        for animal in pool.animalDictionnary.keys():
            #initialisation of a dic for isolated stop
            stopIsolatedDic[animal] = stopIsolatedTimeLine[animal].getDictionary()

            for animalB in pool.animalDictionnary.keys():
                # initialisation of a dic for stop in contact for each individual of the cage
                framesToRemoveFromStopIsolatedTimeLine = []
                stopSocialResult = {}
                if animal == animalB:
                    # print('Same identity')
                    continue
                
                # loop over eventlist
                for stopEvent in stopIsolatedTimeLine[animal].eventList:

                    # for each event we seek in t and search a match in isInContactDictionnary between animal and animalB
                    for t in range ( stopEvent.startFrame, stopEvent.endFrame+1 ) :
                        if t in isInContactSourceDictionnary[(animal, animalB)]:
                            stopSocialResult[t] = True
                            framesToRemoveFromStopIsolatedTimeLine.append(t)                        

                # save stop social at the end of the search for each animal
                # clean the dictionary of the stop events from frames that are overlapping with exclusive contacts
                
                for t in framesToRemoveFromStopIsolatedTimeLine:
                    stopIsolatedDic[animal].pop( t, None )                    
                    
                stopSocialResultTimeLine = EventTimeLine( None, "Stop in contact" , idA=animal , idB=animalB , idC=None , idD=None , loadEvent=False )
                stopSocialResultTimeLine.reBuildWithDictionnary( stopSocialResult )
                stopSocialResultTimeLine.endRebuildEventTimeLine(connection)

            # save stop isolated at the end of the complete process for animalA
            stopIsolatedResultTimeLine = EventTimeLine(None, "Stop isolated", idA=animal, idB=None, idC=None, idD=None, loadEvent=False)
            stopIsolatedResultTimeLine.reBuildWithDictionnary(stopIsolatedDic[animal])
            stopIsolatedResultTimeLine.endRebuildEventTimeLine(connection)



        
    # log process
    from lmtanalysis.TaskLogger import TaskLogger
    t = TaskLogger( connection )
    t.addLog( "Build Event Stop" , tmin=tmin, tmax=tmax )
                       
    # print( "Rebuild event finished." )
    
    
    


'''
Created on 7 sept. 2017
@author: Fab
'''

from lmtanalysis.Detection import *

#matplotlib fix for mac (uncomment if needed)
#matplotlib.use('TkAgg' )

import matplotlib.pyplot as plt
from lmtanalysis.Chronometer import *
import matplotlib as mpl
import numpy as np
import pandas as pd
from collections import defaultdict
from mpl_toolkits.mplot3d import *
import matplotlib.ticker
import math
import time
from lmtanalysis.Measure import *

from statistics import *
#from scipy.spatial import distance
#from scipy.ndimage.measurements import standard_deviation
from statistics import mean
from lmtanalysis.Event import EventTimeLine
from lmtanalysis.Point import Point
from lmtanalysis.Mask import Mask
from lmtanalysis.Util import *
import matplotlib.patches as mpatches
from lxml import etree
import matplotlib.ticker as ticker
#from lmtanalysis.Util import convert_to_d_h_m_s, getDatetimeFromFrame, mute_prints
from pickle import NONE
from matplotlib import colors


idAnimalColor = [ None, "red","green","blue","orange"]

from enum import Enum



def getAnimalColor( animalId ):
    return idAnimalColor[ animalId ]

class Animal():

    def __init__(self, baseId , RFID , name=None, genotype=None , user1 = None, age=None, sex=None, strain=None, setup=None, conn = None ):
        self.baseId = baseId
        self.RFID = RFID
        self.name = name
        self.genotype = genotype
        self.user1 = user1
        self.age = age
        self.sex = sex
        self.strain = strain
        self.setup = setup
        self.conn = conn
        self.detectionDictionnary = {}
        
            
    def setGenotype(self, genotype ):
        self.genotype = genotype
        cursor = self.conn.cursor()
        query = "UPDATE `ANIMAL` SET `GENOTYPE`='{}' WHERE `ID`='{}';".format( genotype, self.baseId )
        cursor.execute( query )
        self.conn.commit()
        cursor.close()
        
        
    def __str__(self):
        return "Animal Id:{id} Name:{name} RFID:{rfid} Genotype:{genotype} User1:{user1}"\
            .format( id=self.baseId, rfid=self.RFID, name=self.name, genotype=self.genotype, user1=self.user1 )

    
    def getColor(self):
        return getAnimalColor( self.baseId )

    
    def getDetectionAt(self, t):
        if t in self.detectionDictionnary:
            return self.detectionDictionnary[t]
        return None

    
    def loadDetection(self, start=None, end=None, lightLoad = False ):
        '''
        lightLoad only loads massX and massY to speed up the load. Then one can only compute basic features such as global speed of the animals
        '''
        # print ( self.__str__(), ": Loading detection.")
        chrono = Chronometer("Load detection")

        # print("IN LOAD DETECTION !!!!!!!")
            
        self.detectionDictionnary.clear()

        cursor = self.conn.cursor()
        query =""
        if lightLoad == True :
            query = "SELECT FRAMENUMBER, MASS_X, MASS_Y FROM DETECTION WHERE ANIMALID={}".format( self.baseId )
        else:
            query = "SELECT FRAMENUMBER, MASS_X, MASS_Y, MASS_Z, FRONT_X, FRONT_Y, FRONT_Z, BACK_X, BACK_Y, BACK_Z,REARING,LOOK_UP,LOOK_DOWN FROM DETECTION WHERE ANIMALID={}".format( self.baseId )

        if ( start != None ):
            query += " AND FRAMENUMBER>={}".format(start )
        if ( end != None ):
            query += " AND FRAMENUMBER<={}".format(end )

        # print( query )
        # print( self.conn )
        cursor.execute( query )

        rows = cursor.fetchall()
        cursor.close()

        for row in rows:
            frameNumber = row[0]
            massX = row[1]
            massY = row[2]
            detection = None
            #filter detection at 0

            if ( massX < 10 ):
                continue

            if not lightLoad:
                massZ = row[3]

                frontX = row[4]
                frontY = row[5]
                frontZ = row[6]

                backX = row[7]
                backY = row[8]
                backZ = row[9]

                rearing = row[10]
                lookUp = row[11]
                lookDown = row[12]

                detection = Detection( massX, massY, massZ, frontX, frontY, frontZ, backX, backY, backZ, rearing, lookUp, lookDown )
            else:
                detection = Detection( massX, massY , lightLoad = True )

            self.detectionDictionnary[frameNumber] = detection

        # print ( self.__str__(), " ", len( self.detectionDictionnary ) , " detections loaded in {} seconds.".format( chrono.getTimeInS( )) )

        
    def loadMask(self , frame ):
        self.setMask( self.getBinaryDetectionMask( frame ) )

        
    def clearMask(self):
        self.setMask( None )

    def getNumberOfDetection(self, tmin, tmax):
        return len ( self.detectionDictionnary.keys() )


    def filterDetectionToKeepOnlyHeadTailDetection(self):
        nbRemoved = 0
        for key in sorted(self.detectionDictionnary.keys()):
            a = self.detectionDictionnary.get( key )
            if not a.isHeadAndTailDetected():
                self.detectionDictionnary.pop( key )
                nbRemoved+=1
        # print( "Filtering head tail detection. number of detection removed:", nbRemoved )

        
    def filterDetectionByInstantSpeed(self , minSpeed, maxSpeed ):
        """
        speed function in LMT use t-1 and t+1 detection to provide a result.
        here we remove spurious tracking jump, so we check on t to t+1 frame.
        speed is in cm per second
        """
        nbRemoved = 0
        for key in sorted(self.detectionDictionnary.keys()):
            a = self.detectionDictionnary.get( key )
            b = self.detectionDictionnary.get( key+1 )

            if ( b==None or a==None):
                continue

            speed = math.hypot( a.massX - b.massX, a.massY - b.massY )*scaleFactor/(1/30)

            if ( speed > maxSpeed or speed < minSpeed ):
                self.detectionDictionnary.pop( key )
                nbRemoved+=1

        # print( "Filtering Instant speed min:",minSpeed, "max:",maxSpeed, "number of detection removed:", nbRemoved )

        
    def filterDetectionByArea(self, x1, y1, x2, y2 ):
        '''
        filter detection in the cage ( using centimeter, starting at top left of the cage )
        '''
        nbRemoved = 0
        for key in sorted(self.detectionDictionnary.keys()):
            a = self.detectionDictionnary.get( key )

            if ( a==None):
                continue

            x = (a.massX - cornerCoordinates50x50Area[0][0] )* scaleFactor
            y = (a.massY - cornerCoordinates50x50Area[0][1] )* scaleFactor

            if ( x < x1 or x > x2 or y < y1 or y > y2 ):
                self.detectionDictionnary.pop( key )
                nbRemoved+=1

        # print( "Filtering area, number of detection removed:", nbRemoved )

        
    def filterDetectionByEventTimeLine( self, eventTimeLineVoc ):
        '''
        filter detection using an event. Keep only what matches the event
        '''
        eventDic = eventTimeLineVoc.getDictionnary()
        nbRemoved = 0
        for key in sorted( self.detectionDictionnary.keys() ):
            a = self.detectionDictionnary.get( key )

            if ( a==None):
                continue

            if not ( key in eventDic ):
                self.detectionDictionnary.pop( key )
                nbRemoved+=1

        # print( "Filtering area, number of detection removed:", nbRemoved )

        
    def clearDetection(self):
        self.detectionDictionnary.clear()

        
    def getMaxDetectionT(self):
        """
        returns the timepoint of the last detection.
        """

        if len ( self.detectionDictionnary.keys() ) == 0:
            return None

        return sorted(self.detectionDictionnary.keys())[-1]
    
    
    def getTrajectoryData( self , maskingEventTimeLine=None ):
        keyList = sorted(self.detectionDictionnary.keys())

        if maskingEventTimeLine!=None:
            keyList = maskingEventTimeLine.getDictionnary()

        xList = []
        yList = []

        previousKey = 0

        for key in keyList:
            #print ( "key:", key, "value", self.getSpeed( key ) , "previous:" , previousKey )

            if previousKey+1 != key:
                xList.append( [np.nan, np.nan] )
                yList.append( [np.nan, np.nan] )
                previousKey = key
                #print("break previous")
                continue
            previousKey = key

            a = self.detectionDictionnary.get( key )
            if a == None:
                xList.append( [np.nan, np.nan] )
                yList.append( [np.nan, np.nan] )
                #print("break none A")
                continue
            b = self.detectionDictionnary.get( key+1 )
            if b == None:
                xList.append( [np.nan, np.nan] )
                yList.append( [np.nan, np.nan] )
                #print("break none B")
                continue

            xList.append( [a.massX,b.massX] )
            yList.append( [-a.massY,-b.massY] )
            
        # print(xList)
        # print(' //// ')
        # print(yList)

        return xList, yList

    
    def getNoseTrajectoryData( self , maskingEventTimeLine=None ):
        keyList = sorted(self.detectionDictionnary.keys())

        if maskingEventTimeLine!=None:
            keyList = maskingEventTimeLine.getDictionnary()

        xList = []
        yList = []

        previousKey = 0


        for key in keyList:

            #print ( "key:", key, "value", self.getSpeed( key ) , "previous:" , previousKey )

            if previousKey+1 != key:
                xList.append( [np.nan, np.nan] )
                yList.append( [np.nan, np.nan] )
                previousKey = key

                #print("break previous")
                continue
            previousKey = key

            a = self.detectionDictionnary.get( key )
            if ( a==None):
                xList.append( [np.nan, np.nan] )
                yList.append( [np.nan, np.nan] )
                #print("break none A")
                continue
            b = self.detectionDictionnary.get( key+1 )
            if ( b==None):
                xList.append( [np.nan, np.nan] )
                yList.append( [np.nan, np.nan] )
                #print("break none B")
                continue

            xa = a.frontX
            xb = b.frontX
            ya = -a.frontY
            yb = -b.frontY

            if xa < 0:
                xa = np.nan
                xb = np.nan
                ya = np.nan
                yb = np.nan

            xList.append( [xa,xb] )
            yList.append( [ya,yb] )

        return xList, yList

    
    def plotTrajectory(self , show=True, color='k', maskingEventTimeLine=None, title = "" ):
        print ("(within Animal) Draw trajectory of animal " + self.name )

        xList, yList = self.getTrajectoryData( maskingEventTimeLine )

        plt.plot( xList, yList, color=color, linestyle='-', linewidth=1, alpha=0.5, label= self.name )
        plt.title( title + self.RFID )
        plt.xlim(90, 420)
        plt.ylim(-370, -40)

        if ( show ):
            plt.show()
            
    def plotHeatmap(self , show=True, color='k', maskingEventTimeLine=None, title = "" ):
        print ("(within Animal) Draw Heatmap of animal " + self.name )

        xList, yList = self.getTrajectoryData( maskingEventTimeLine )

        plt.hist2d( xList, yList, color=color, linestyle='-', linewidth=1, alpha=0.5, label= self.name )
        plt.title( title + self.RFID )
        plt.xlim(90, 420)
        plt.ylim(-370, -40)

        if show:
            plt.show()


    def plotTrajectory3D(self):

        print ("Draw 3D trajectory")
        keyList = sorted(self.detectionDictionnary.keys())

        mpl.rcParams['legend.fontsize'] = 10

        fig = plt.figure()
        ax = fig.gca(projection='3d')
        theta = np.linspace(-4 * np.pi, 4 * np.pi, 100)

        z = np.linspace(-2, 2, 100)
        r = z**2 + 1
        x = r * np.sin(theta)
        y = r * np.cos(theta)

        '''
        print ( z )
        print ( r )
        print ( x )
        print ( y )
        '''

        xList = []
        yList = []
        zList = []

        for key in keyList:

            a = self.detectionDictionnary.get( key )
            b = self.detectionDictionnary.get( key+1 )
            if ( b==None):
                continue

            xList.append( a.massX )
            yList.append( a.massY )
            zList.append( a.massZ )

        ax.plot(xList, yList, zList, label= "3D Trajectory of " + self.RFID )
        ax.legend()

        plt.show()


    def getDistancePerBin(self , binFrameSize , minFrame=0, maxFrame=None ):
        if ( maxFrame==None ):
            maxFrame= self.getMaxDetectionT()

        distanceList = []
        t = minFrame
        while ( t < maxFrame ):
            distanceBin = self.getDistance( t , t+binFrameSize )
            # print( "Distance bin n:{} value:{}".format ( t , distanceBin ) )
            distanceList.append( distanceBin )
            t=t+binFrameSize

        return distanceList


    def getDistance(self, tmin=0, tmax=None):
        '''
        Returns the distance traveled by the animal (in cm)
        '''
        # print("Compute total distance min:{} max:{} ".format(tmin, tmax))


        '''
        keyList = list(self.detectionDictionnary.keys())
        if not alreadySorted:
            keyList = sorted(self.detectionDictionnary.keys())
        '''
        if (tmax==None):
            tmax= self.getMaxDetectionT()

        totalDistance = 0
        for t in range(tmin , tmax):
        #for key in keyList:
            '''
            if ( key <= tmin or key >= tmax ):
                continue
            '''

            a = self.detectionDictionnary.get(t)
            b = self.detectionDictionnary.get(t+1)

            if b==None or a==None:
                continue
            distance = math.hypot( a.massX - b.massX, a.massY - b.massY )
            #print('distance: ', distance)
            if ( distance >85.5): #if the distance calculated between two frames is too large, discard
                continue

            totalDistance += distance

        totalDistance *= scaleFactor

        return totalDistance


    def getOrientationVector(self, t):

        d = self.detectionDictionnary.get( t )

        if d == None:
            return None

        if d.frontX == None:
            return None

        if d.backX == None:
            return None

        deltaX = d.frontX - d.backX
        deltaY = d.frontY - d.backY
        p = Point( deltaX, deltaY )
        return p

    def getSpeedVector(self, t):

        a = self.detectionDictionnary.get( t-1 )
        b = self.detectionDictionnary.get( t+1 )

        if a == None or b == None:
            return None

        speedVectorX = a.massX - b.massX
        speedVectorY = a.massY - b.massY

        p = Point( speedVectorX, speedVectorY )
        return p

    def getFrontSpeed(self, t):

        a = self.detectionDictionnary.get( t-1 )
        b = self.detectionDictionnary.get( t+1 )

        if a == None or b == None:
            return None

        speedVectorX = a.frontX - b.frontX
        speedVectorY = a.frontY - b.frontY

        p = Point( speedVectorX, speedVectorY )
        return p

    def getBackSpeed(self, t):

        a = self.detectionDictionnary.get( t-1 )
        b = self.detectionDictionnary.get( t+1 )

        if a == None or b == None:
            return None

        speedVectorX = a.backX - b.backX
        speedVectorY = a.backY - b.backY

        p = Point( speedVectorX, speedVectorY )
        return p


    def getDistanceSpecZone(self, tmin=0, tmax=None, xa=None, ya=None, xb=None, yb=None):

        keyList = sorted(self.detectionDictionnary.keys()) #get the list of the frames where the animal has been detected

        if ( tmax==None ):
            tmax = self.getMaxDetectionT()

        distance = 0 #initialise the distance

        for key in keyList:

            if ( key <= tmin or key >= tmax ):
                #print ( 1 )
                continue

            a = self.detectionDictionnary.get( key ) #get the detection of the animal at the given frame
            b = self.detectionDictionnary.get( key+1 ) #get the detection of the animal at the following frame

            if ( b==None):
                continue

            if (a.massX<xa or a.massX>xb or a.massY<ya or a.massY>yb or b.massX<xa or b.massX>xb or b.massY<ya or b.massY>yb):
                #if the animal is not in the zone, then the distance is not computed
                continue

            if (math.hypot( a.massX - b.massX, a.massY - b.massY )>85.5): #if the distance calculated between two frames is too large, discard
                continue

            distance += math.hypot( a.massX - b.massX, a.massY - b.massY ) #add the distance to the previously calculated distance

        distance *= scaleFactor #convert the distance in cm

        return distance


    def getDistanceTo (self, t, animalB):
        '''
        determine the distance between the focal animal and another one specified in argument at one specified time point t
        check before that both animals are detected at this time point
        '''
        distanceTo = None

        if ( not ( t in animalB.detectionDictionnary ) ):
            return None

        if ( not ( t in self.detectionDictionnary ) ):
            return None

        if (animalB.detectionDictionnary[t].massX == None):
            return None

        if (math.hypot( self.detectionDictionnary[t].massX - animalB.detectionDictionnary[t].massX, self.detectionDictionnary[t].massY - animalB.detectionDictionnary[t].massY ) > 71*57/10): #if the distance calculated between the two individuals is too large, discard
            return None

        else:
            distanceTo = math.hypot( self.detectionDictionnary[t].massX - animalB.detectionDictionnary[t].massX, self.detectionDictionnary[t].massY - animalB.detectionDictionnary[t].massY )
            return distanceTo


    def getDistanceToPoint (self, t, xPoint, yPoint):
        '''
        determine the distance between the focal animal and a specific point in the arena at one specified time point t
        '''
        distanceToPoint = None

        if ( not ( t in self.detectionDictionnary ) ):
            return None

        if (math.hypot( self.detectionDictionnary[t].massX - xPoint, self.detectionDictionnary[t].massY - yPoint ) > MAX_DISTANCE_THRESHOLD): #if the distance calculated is too large, discard
            return None

        else:
            distanceToPoint = math.hypot( self.detectionDictionnary[t].massX - xPoint, self.detectionDictionnary[t].massY - yPoint )
            return distanceToPoint

    def getDistanceNoseToPoint (self, t, xPoint, yPoint):
        '''
        determine the distance between the nose of the animal and a specific point in the arena at one specified time point t
        '''
        distanceNoseToPoint = None

        if ( not ( t in self.detectionDictionnary ) ):
            return None
        if (self.detectionDictionnary[t].frontX < 0):
            return None
        if (math.hypot( self.detectionDictionnary[t].massX - xPoint, self.detectionDictionnary[t].massY - yPoint ) > MAX_DISTANCE_THRESHOLD): #if the distance calculated is too large, discard
            return None

        else:
            distanceNoseToPoint = math.hypot( self.detectionDictionnary[t].frontX - xPoint, self.detectionDictionnary[t].frontY - yPoint )
            return distanceNoseToPoint


    def getMeanBodyLength (self, tmin=0, tmax=None):
        '''
        determine the mean body length over the time window specified
        '''
        keyList = sorted(self.detectionDictionnary.keys())

        if ( tmax==None ):
            tmax= self.getMaxDetectionT()

        if ( tmax==None ):
            self.meanBodyLength = None
            return self.meanBodyLength

        bodySizeList = []

        for key in keyList:

            if ( key <= tmin or key >= tmax ):
                continue

            a = self.detectionDictionnary.get( key )
            
            if (a.isHeadAndTailDetected()):
                bodySizeList.append(a.getBodySize())

        mean = np.nanmean(bodySizeList)
        # print( "mean animal bodysize: "  , mean )

        self.meanBodyLength = mean

        return self.meanBodyLength


    def getBodyThreshold (self, tmin=0, tmax=None):
        '''
        determine the body size threshold used to determine SAP
        '''
        keyList = sorted(self.detectionDictionnary.keys())

        if ( tmax==None ):
            tmax= self.getMaxDetectionT()

        bodySizeList = []

        for key in keyList:

            if ( key <= tmin or key >= tmax ):
                continue

            a = self.detectionDictionnary.get( key )
            bodySizeList.append(a.getBodySize())

        #verifier si ce calcul tourne bien:
        threshold = np.nanmean(bodySizeList) + np.nanstd(bodySizeList)

        self.bodyThreshold = threshold

        return self.bodyThreshold


    def getMedianBodyHeight (self, tmin=0, tmax=None):
        '''
        determine the body size threshold used to determine SAP
        '''
        keyList = sorted(self.detectionDictionnary.keys())

        if ( tmax==None ):
            tmax= self.getMaxDetectionT()

        bodyHeightList = []

        for key in keyList:

            if ( key <= tmin or key >= tmax ):
                continue

            a = self.detectionDictionnary.get( key )
            bodyHeightList.append(a.massZ)

        self.medianBodyHeight = np.median(np.array(bodyHeightList))

        return self.medianBodyHeight


    def getThresholdMassHeight (self, tmin=0, tmax=None):
        '''
        determine the body size height threshold used to determine whether the animal is rearing or not
        here we use the 7th decile
        '''
        keyList = sorted(self.detectionDictionnary.keys())

        if ( tmax==None ):
            tmax= self.getMaxDetectionT()

        massHeightList = []

        for key in keyList:

            if ( key <= tmin or key >= tmax ):
                continue

            a = self.detectionDictionnary.get( key )
            massHeightList.append(a.massZ)

        decile = 7*len(massHeightList)/10
        # print("7eme decile mass:")
        # print(decile)
        # print("arrondi massZ:")
        # print(math.ceil(decile)-1)

        self.eightDecileMassHeight = sorted(massHeightList)[math.ceil(decile)-1]

        return self.eightDecileMassHeight


    def getThresholdFrontHeight (self, tmin=0, tmax=None):
        '''
        determine the body size height threshold used to determine whether the animal is rearing or not
        here we use the 7th decile
        '''
        keyList = sorted(self.detectionDictionnary.keys())

        if ( tmax==None ):
            tmax= self.getMaxDetectionT()

        frontHeightList = []

        for key in keyList:

            if ( key <= tmin or key >= tmax ):
                continue

            a = self.detectionDictionnary.get( key )
            frontHeightList.append(a.frontZ)

        decile = 7*len(frontHeightList)/10
        # print("7eme decile front:")
        # print(decile)
        # print("arrondi frontZ:")
        # print(math.ceil(decile)-1)

        self.eightDecileFrontHeight = sorted(frontHeightList)[math.ceil(decile)-1]

        return self.eightDecileFrontHeight


    def getDirection(self, t):
        '''
        determines the direction of the animal using the head and the mass center
        '''
        a = self.detectionDictionnary.get( t )
        return a.getDirection();


    def getSpeed(self , t ):
        '''
        calculate the instantaneous speed of the animal at each frame
        '''
        a = self.detectionDictionnary.get( t-1 )
        b = self.detectionDictionnary.get( t+1 )

        if ( b==None or a==None):
            return None

        speed = math.hypot( a.massX - b.massX, a.massY - b.massY )*scaleFactor/(2/30)
        

        return speed
    
    
    def getSpeedOverTimePeriod(self, tmin, tmax):
        '''Compute the speed of the animal over a time period'''          
        duration = tmax - tmin
        sum = 0
        speedList = []
        
        for t in range ( tmin, tmax+1 ) :
            speed = self.getSpeed(t)
            #print('speed: ', speed)
            if ( speed != None ):
                sum+= speed
                speedList.append( speed )
        
        if speedList != []:
            meanSpeed = sum / duration
            maxSpeed = max(speedList)
            minSpeed = min(speedList)
        else:
            meanSpeed = None
            maxSpeed = None
            minSpeed = None
        
        return ( duration, meanSpeed, minSpeed, maxSpeed )


    def getVerticalSpeed(self , t ):
        '''
        calculate the instantaneous vertical speed of the mass center of the animal at each frame
        '''
        a = self.detectionDictionnary.get( t-1 )
        b = self.detectionDictionnary.get( t+1 )

        if ( b==None or a==None):
            return None

        verticalSpeed = (b.massZ - a.massZ)/2

        return verticalSpeed


    def getSap(self, tmin=0, tmax=None, xa=None, ya=None, xb=None, yb=None):
        # print("Compute number of frames in SAP in specific zone of the cage")

        self.getBodyThreshold( )
        self.getMedianBodyHeight()

        #TODO: pas mettre body threshold en self car ... c'est pas la peine.

        keyList = sorted(self.detectionDictionnary.keys())

        if ( tmax==None ):
            tmax = self.getMaxDetectionT()

        sapList = []

        for key in keyList:

            if ( key <= tmin or key >= tmax ):
                #print ( 1 )
                continue

            detection = self.detectionDictionnary.get(key);
            if (detection.massX<xa or detection.massX>xb or detection.massY<ya or detection.massY>yb or detection.massZ>self.medianBodyHeight):
                #print ( 2 )
                continue

            speed = self.getSpeed( key )
            if ( speed == None ):
                #print ( 3 )
                continue

            if (detection.getBodySize( ) >=self.bodyThreshold and speed<SPEED_THRESHOLD_LOW and detection.massZ<self.medianBodyHeight):
                #print ( 4 )
                sapList.append( detection )

        return sapList


    def getSapDictionnary(self, tmin=0, tmax=None ):

        self.getBodyThreshold( )
        self.getMedianBodyHeight()

        keyList = sorted(self.detectionDictionnary.keys())

        if ( tmax==None ):
            tmax = self.getMaxDetectionT()

        sapDictionnary = {}

        for key in keyList:

            if ( key <= tmin or key >= tmax ):
                continue

            detection = self.detectionDictionnary.get(key);

            speed = self.getSpeed( key )
            if ( speed == None ):
                continue

            if (detection.getBodySize( ) >=self.bodyThreshold and speed<SPEED_THRESHOLD_LOW and detection.massZ<self.medianBodyHeight):
                sapDictionnary[key] = True

        return sapDictionnary


    def getCountFramesSpecZone(self, tmin=0, tmax=None, xa=None, ya=None, xb=None, yb=None):
        '''
        coordinates are in pixel
        '''
        keyList = sorted(self.detectionDictionnary.keys())

        x1=min( xa, xb )
        x2=max( xa, xb )
        y1=min( ya, yb )
        y2=max( ya, yb )

        if ( tmax==None ):
            tmax = self.getMaxDetectionT()

        count =0

        for key in keyList:

            if ( key <= tmin or key >= tmax ):
                #print ( 1 )
                continue

            detection = self.detectionDictionnary.get(key);
            if (detection.massX<x1 or detection.massX>x2 or detection.massY<y1 or detection.massY>y2):
                #print ( 2 )
                continue

            count+=1

        return count


    def plotDistance(self, color='k' , show=True ):
        print ("Plot distance")
        keyList = sorted(self.detectionDictionnary.keys())

        tList = []
        distanceList = []

        totalDistance = 0
        for key in keyList:

            a = self.detectionDictionnary.get( key )
            b = self.detectionDictionnary.get( key+1 )

            if ( b==None):
                continue

            totalDistance += math.hypot( a.massX - b.massX, a.massY - b.massY )

            tList.append( key / 30 / 60 )
            distanceList.append( totalDistance )

        #fig,ax = plt.subplots()
        plt.plot( tList, distanceList, color=color, linestyle='-', linewidth=2 , label="Cumulated distance of " + self.__str__() )

        #formatter = matplotlib.ticker.FuncFormatter(lambda frame, x: time.strftime('%Hh%Mm%Ss', time.gmtime( frame // 30)))
        #formatter = matplotlib.ticker.FuncFormatter(lambda frame, x: time.strftime('%Hh%Mm', time.gmtime( frame // 30)))
        #ax.xaxis.set_major_formatter( formatter )
        #ax.xaxis.set_minor_locator( matplotlib.ticker.MultipleLocator(30*60) )
        #ax.xaxis.set_major_locator( matplotlib.ticker.MultipleLocator(30*60*10) )

        plt.legend()

        if ( show ):
            plt.show()


    def getBinaryDetectionMask(self , t):
        '''
        returns the mask of a detection at a given T.
        '''
        query = "SELECT DATA FROM DETECTION WHERE ANIMALID={} AND FRAMENUMBER={}".format( self.baseId , t )

        cursor = self.conn.cursor()
        cursor.execute( query )

        rows = cursor.fetchall()
        cursor.close()

        if ( len(rows)!=1 ):
            #print ("unexpected number of row: " , str( len(rows ) ) )
            return None

        row = rows[0]
        data = row[0]

        mask = Mask( data, self.getColor( ) )

        return mask


class AnimalPool():
    """
    Manages an experiment.
    """

    def __init__(self):

        self.animalDictionnary = {}
        self.detectionStartFrame = None
        self.detectionEndFrame   = None

    def getAnimalDictionnary(self):
        return self.animalDictionnary

    def getAnimalWithId(self , id):
        return self.animalDictionnary[id]

    def getAnimalList(self):

        animalList= []

        for k in self.animalDictionnary:
            animal = self.animalDictionnary[k]
            animalList.append( animal )

        return animalList

    def loadAnimals( self, conn ):

        # print ("Loading animals.")

        cursor = conn.cursor()
        self.conn = conn

        # check experiment parameters
        
        
        '''
        try:
            print( "Checking animal type.")
            query = "SELECT * FROM EXPERIMENT"
            print("2")
            cursor.execute( query )
            print("3")
            rows = cursor.fetchall()
            print("4")
            print( rows )
            for row in rows:
                print("test -- sdlqsfhldjk")
                print( str(row[1]).lower() )
                if ( "rat" in str(row[1]).lower() ):
                    animalType = AnimalType.RAT
                    
                if ( "mouse" in str(row[1]).lower() ):
                    animalType = AnimalType.RAT
        
            
        except:
            print("Failed to load experiment data. Mouse by default.")
        '''

        # Check the number of row available in base
        query = "SELECT * FROM ANIMAL"
        cursor.execute( query )
        field_names = [i[0] for i in cursor.description]
        # print ( "Fields available in lmtanalysis: " , field_names )

        #build query
        query = "SELECT "
        nbField = len( field_names )
        if ( nbField == 3 ):
            query+="ID,RFID,NAME"
        elif ( nbField == 4 ):
            query+="ID,RFID,NAME,GENOTYPE"
        elif ( nbField == 5 ):
            query+="ID,RFID,NAME,GENOTYPE,SEX"
        elif ( nbField == 7 ):
            query+="ID,RFID,NAME,GENOTYPE,AGE,SEX,STRAIN"
        elif ( nbField == 8 ):
            query+="ID,RFID,NAME,GENOTYPE,AGE,SEX,STRAIN,SETUP"
        elif ( nbField == 9 ):
            query+="ID,RFID,NAME,GENOTYPE,AGE,SEX,STRAIN,SETUP,IND"

        query += " FROM ANIMAL ORDER BY GENOTYPE"
        # print ( "SQL Query: " + query )

        cursor.execute( query )
        rows = cursor.fetchall()
        cursor.close()

        self.animalDictionnary.clear()

        for row in rows:

            animal = None
            if ( len( row ) == 3 ):
                animal = Animal( row[0] , row[1] , name=row[2] , conn = conn )
            if ( len( row ) == 4 ):
                animal = Animal( row[0] , row[1] , name=row[2] , genotype=row[3] , conn = conn )
            if ( len( row ) == 5 ):
                animal = Animal( row[0] , row[1] , name=row[2] , genotype=row[3] , user1=row[4] , conn = conn )
            if ( len( row ) == 7 ):
                animal = Animal( row[0] , row[1] , name=row[2] , genotype=row[3] , age=row[4] , sex=row[5] , strain=row[6], conn = conn )
            if ( len( row ) == 8 ):
                animal = Animal( row[0] , row[1] , name=row[2] , genotype=row[3] , age=row[4] , sex=row[5] , strain=row[6], setup=row[7], conn = conn )
            if ( len( row ) == 9 ):
                animal = Animal( row[0] , row[1] , name=row[2] , genotype=row[3] , age=row[4] , sex=row[5] , strain=row[6], setup=row[7], user1=row[8], conn = conn )

            if ( animal!= None):
                self.animalDictionnary[animal.baseId] = animal
                # print ( animal )
            else:
                continue
                # print ( "Animal loader : error while loading animal.")

    def getAnonymousDetection(self, frame ):
        if frame not in self.anonymousDetection:
            return None

        return self.anonymousDetection[frame]

    def loadAnonymousDetection(self, start = None, end=None ):
        '''
        Load the dictionary of anonymous detection.
        each entry get a list of detection
        clear previous anonymous detection dictionary
        '''
        self.anonymousDetection = {}

        chrono = Chronometer("Load anonymous detection")
        cursor = self.conn.cursor()

        #query = "SELECT FRAMENUMBER, MASS_X, MASS_Y FROM DETECTION WHERE ANIMALID IS NULL"
        query = "SELECT FRAMENUMBER, MASS_X, MASS_Y FROM DETECTION WHERE ANIMALID IS NULL"

        if ( start != None ):
            query += " AND FRAMENUMBER>={}".format(start )
        if ( end != None ):
            query += " AND FRAMENUMBER<={}".format(end )

        # print( query )
        cursor.execute( query )

        rows = cursor.fetchall()
        cursor.close()

        for row in rows:
            frameNumber = row[0]
            massX = row[1]
            massY = row[2]
            #data = row[3]
            detection = None
            #filter detection at 0
            if ( massX < 10 ):
                continue

            detection = Detection( massX, massY , lightLoad = True )
            #detection.setMask( Mask( data ) )

            if frameNumber not in self.anonymousDetection:
                self.anonymousDetection[frameNumber] = []

            self.anonymousDetection[frameNumber].append( detection )

        # print ( len( self.anonymousDetection ) , " frames containing anonymous detections loaded in {} seconds.".format( chrono.getTimeInS( )) )

    def loadDetection (self , start = None, end=None , lightLoad = False ):
        self.detectionStartFrame = start
        self.detectionEndFrame   = end
        for animal in self.animalDictionnary.keys():
            self.animalDictionnary[animal].loadDetection( start = start, end = end , lightLoad=lightLoad )

    def filterDetectionByInstantSpeed(self, minSpeed, maxSpeed):
        for animal in self.animalDictionnary.keys():
            self.animalDictionnary[animal].filterDetectionByInstantSpeed( minSpeed, maxSpeed )

    def filterDetectionToKeepOnlyHeadTailDetection(self):
        for animal in self.animalDictionnary.keys():
            self.animalDictionnary[animal].filterDetectionToKeepOnlyHeadTailDetection( )

    def filterDetectionByArea(self, x1, y1, x2, y2 ):
        for animal in self.animalDictionnary.keys():
            self.animalDictionnary[animal].filterDetectionByArea( x1, y1, x2, y2 )

    def filterDetectionByEventTimeLine(self, eventTimeLineVoc ):
        for animal in self.animalDictionnary.keys():
            self.animalDictionnary[animal].filterDetectionByEventTimeLine( eventTimeLineVoc )

    def getGenotypeList(self):

        genotype = {}

        for k in self.animalDictionnary:
            animal = self.animalDictionnary[k]
            genotype[animal.genotype] = True

        return genotype.keys()

    def getAnimalsWithGenotype(self, genotype ):

        resultList= []

        for k in self.animalDictionnary:
            animal = self.animalDictionnary[k]
            if ( animal.genotype == genotype ):
                resultList.append( animal )

        return resultList

        #return [x for x in self.animalDictionnary if x.genotype==genotype ]

    def getNbAnimals(self):
        return len(self.animalDictionnary)


    def getMaxDetectionT(self):
        """
        returns the timepoint of the last detection of all animals
        """
        maxFrame = 0
        for animal in self.getAnimalList():
            maxFrame = max( maxFrame, animal.getMaxDetectionT() )

        return maxFrame

    def frameToTimeTicker(self, x, pos):
        vals= convert_to_d_h_m_s( x )
        experimentTime = "D{0} - {1:02d}:{2:02d}".format( int(vals[0])+1, int(vals[1]), int(vals[2]) )

        realTime = ""
        if x > 0:
            datetime = getDatetimeFromFrame( self.conn , x )
            if datetime != None:
                realTime = "\n" + getDatetimeFromFrame( self.conn , x ).strftime('%d-%b-%Y %H:%M:%S')

        return experimentTime + realTime

    def plotSensorData( self , sensor="TEMPERATURE", show = True , saveFile = None , minValue=0 , autoNight= False):
        """
        plots data for temperature
        """
        # print( "plotting sensor data. " , sensor )
        cursor = self.conn.cursor()

        # Check the number of row available in base
        query = "SELECT FRAMENUMBER,"+sensor+" FROM FRAME"
        try:
            cursor.execute( query )
        except:
            # print("plot sensor data: can't access data for ", sensor )
            return

        rows = cursor.fetchall()
        cursor.close()

        frameNumberList = []
        sensorValueList = []
        for row in rows:

            sensorValue = row[1]
            if ( sensorValue > minValue ):
                if sensorValue != None and minValue !=None:
                    frameNumberList.append( row[0] )
                    sensorValueList.append( sensorValue )

        fig, ax = plt.subplots( nrows = 1 , ncols = 1 , figsize=( 24, 8 ) )

        ''' set x axis '''
        formatter = ticker.FuncFormatter( self.frameToTimeTicker )
        ax.xaxis.set_major_formatter(formatter)

        ax.tick_params(labelsize=20 )
        ax.xaxis.set_major_locator(ticker.MultipleLocator( 30 * 60 * 60 * 12 ))
        ax.xaxis.set_minor_locator(ticker.MultipleLocator( 30 * 60 * 60 ))

        ax.plot( frameNumberList, sensorValueList, color= "black" , linestyle='-', linewidth=1, label= sensor )
        ax.legend( loc=1 )

        # plot night
        nightTimeLine = EventTimeLine( self.conn, "night" , None, None, None , None )

        mean = 0
        std = 0

        if len( sensorValueList ) > 1 :
            mean = np.mean( sensorValueList )
            std = np.std( sensorValueList )

        for nightEvent in nightTimeLine.eventList:

            ax.axvspan( nightEvent.startFrame, nightEvent.endFrame, alpha=0.1, color='black', ymin= 0, ymax = 1 )
            ax.text( (nightEvent.startFrame+ nightEvent.endFrame)/2 , mean+std , "dark phase" ,fontsize=20,ha='center')

        autoNightList = []
        if autoNight:
            # compute mean value:
            plt.axhline(y=mean, color='r', linestyle='--' )
            ax.text( 0 , mean , "auto night threshold" ,fontsize=20, ha='left' , va='bottom')

            inNight=False
            start = None
            end = None
            for i in range(len(frameNumberList)):
                frame = frameNumberList[i]
                value = sensorValueList[i]
                if value < mean: # night
                    if inNight:
                        continue
                    else:
                        inNight = True
                        start=frame
                else:
                    if inNight:
                        end = frame
                        inNight = False
                        if end-start > 300:
                            autoNightList.append( ( start, end ) )
                        else:
                            continue
                            # print("Skipping very short night phase. (less than 10 seconds)")
                        start = None
                        end = None

            for autoNight in autoNightList:

                ax.axvspan( autoNight[0], autoNight[1], alpha=0.1, color='red', ymin= 0.2, ymax = 0.8 )
                ax.text( (autoNight[0]+autoNight[1])/2 , mean , "auto dark phase" ,fontsize=20,ha='center', color='red')

        plt.xlabel('time')

        if sensor == "TEMPERATURE":
            plt.ylabel('Temperature (C)')
            plt.ylim( 17, 28 )


        if sensor == "SOUND":
            plt.ylabel('Sound level (indice)')

        if sensor == "HUMIDITY":
            plt.ylabel('Humidity %')
            plt.ylim( 0, 100 )

        if sensor == "LIGHTVISIBLE":
            plt.ylabel('Visible light (indice)')

        if sensor == "LIGHTVISIBLEANDIR":
            plt.ylabel('Visible and infrared light (indice)')

        if saveFile !=None:
            # print("Saving figure : " + saveFile )
            fig.savefig( saveFile, dpi=100)

        '''
        print("TEST 1")
        import os
        os.startfile( saveFile, 'open')
        print( "TEST 2")
        '''

        if ( show ):
            plt.show()
        plt.close()

        if autoNight:
            return autoNightList


    def plotNight(self, show = True , saveFile = None ):

        fig, ax = plt.subplots( nrows = 1 , ncols = 1 , figsize=( 12, 4 ) )

        nightTimeLine = EventTimeLine( self.conn, "night" , None, None, None , None )
        for nightEvent in nightTimeLine.eventList:

            ax.axvspan( nightEvent.startFrame, nightEvent.endFrame, alpha=0.1, color='black')
            ax.text( (nightEvent.startFrame+ nightEvent.endFrame)/2 , 0.5 , "dark phase" ,fontsize=8,ha='center')

        ''' set x axis '''
        formatter = ticker.FuncFormatter( self.frameToTimeTicker )
        ax.xaxis.set_major_formatter(formatter)
        ax.tick_params(labelsize=6 )
        ax.xaxis.set_major_locator(ticker.MultipleLocator( 30 * 60 * 60 * 12 ))
        ax.xaxis.set_minor_locator(ticker.MultipleLocator( 30 * 60 * 60 ))

        if saveFile !=None:
            # print("Saving figure : " + saveFile )
            fig.savefig( saveFile, dpi=100)

        if ( show ):
            plt.show()
        plt.close()

    def buildSensorData(self, file , show=False ):
        # print("Build sensor data")
        #self.plotNight( show = show, saveFile = file+"_day night.pdf" )
        self.plotSensorData( sensor = "TEMPERATURE" , minValue = 10, saveFile = file+"_log_temperature.pdf", show = show )
        self.plotSensorData( sensor = "SOUND" , saveFile = file+"_log_sound level.pdf" , show = show )
        self.plotSensorData( sensor = "HUMIDITY" , minValue = 5 , saveFile = file+"_log_humidity.pdf" ,show = show )
        self.plotSensorData( sensor = "LIGHTVISIBLE" , minValue = 40 , saveFile = file+"_log_light visible.pdf", show = show  )
        self.plotSensorData( sensor = "LIGHTVISIBLEANDIR" , minValue = 50 , saveFile = file+"_log_light visible and infra.pdf", show = show  )

    def plotTrajectory( self , show=True, maskingEventTimeLine=None , title = None, scatter = False, saveFile = None ):
        # print( "AnimalPool: plot trajectory.")
        nbCols = len( self.getAnimalList() )+1
        fig, axes = plt.subplots( nrows = 1 , ncols = nbCols , figsize=( nbCols*4, 1*4 ) , sharex='all', sharey='all'  )

        if title==None:
            title="Trajectory of animals"

        # Draw Trajectories of all animals in one graph
        axis = axes[0]
        legendList=[]
        for animal in self.getAnimalList():
            # print("for loop #1")
            # print ("Compute trajectory of animal " + animal.name )
            xList, yList = animal.getTrajectoryData( maskingEventTimeLine )
            
            # print ("Draw trajectory of animal " + animal.name )
            if scatter == True:
                axis.scatter( xList, yList, color= animal.getColor() , s=1 , linewidth=1, alpha=0.05, label= animal.RFID )
                legendList.append( mpatches.Patch(color=animal.getColor(), label=animal.RFID) )
            else:
                axis.plot( xList, yList, color= animal.getColor() , linestyle='-', linewidth=1, alpha=0.5, label= animal.RFID )

        axis.legend( handles = legendList , loc=1 )
        axis.set_xlim(90, 420)
        axis.set_ylim(-370, -40)

        # draw separated animals in separated sub-graphs
        for animal in self.getAnimalList():
            # print("for loop #2")
            axis = axes[self.getAnimalList().index(animal)+1]

            legendList=[]

            # print ("Compute trajectory of animal " + animal.name )
            xList, yList = animal.getTrajectoryData( maskingEventTimeLine )
            
            # print ("Draw trajectory of animal " + animal.name )
            if scatter == True:
                axis.scatter( xList, yList, color= animal.getColor() , s=1 , linewidth=1, alpha=0.05, label= animal.RFID )
                legendList.append( mpatches.Patch(color=animal.getColor(), label=animal.RFID) )
            else:
                axis.plot( xList, yList, color= animal.getColor() , linestyle='-', linewidth=1, alpha=0.5, label= animal.RFID )

            axis.legend( handles = legendList , loc=1 )

        fig.suptitle( title )

        if saveFile !=None:
            # print("Saving figure : " + saveFile )
            fig.savefig( saveFile, dpi=200)

        if show:
            plt.show()
            
        plt.close()

        
    def plotHeatmap( self , show=True, maskingEventTimeLine=None , title = None, scatter = False, saveFile = None ):
        # print( "AnimalPool: plot Heatmap.")
        nbCols = len( self.getAnimalList() )+1
        fig, axes = plt.subplots( nrows = 1 , ncols = nbCols , figsize=( nbCols*4, 1*4 ) , sharex='all', sharey='all'  )

        if title==None:
            title="Heatmap of animals"

        # Draw all animals together
        axis = axes[0]
        legendList=[]
        for animal in self.getAnimalList():

            # print ("Compute trajectory of animal " + animal.name )
            xList2, yList2 = animal.getTrajectoryData( maskingEventTimeLine )
                        
            # if scatter == True:
            #     axis.scatter( xList, yList, color= animal.getColor() , s=1 , linewidth=1, alpha=0.05, label= animal.RFID )
            #     legendList.append( mpatches.Patch(color=animal.getColor(), label=animal.RFID) )
            # else:
            #     axis.hist2d( xList, yList)
            
            xList_singlevalue = [item[0] for item in xList2]
            yList_singlevalue = [item[0] for item in yList2]
            
            new_xlist = [item for item in xList_singlevalue if not(math.isnan(item)) == True]
            new_ylist = [item for item in yList_singlevalue if not(math.isnan(item)) == True]
            
            xmin = min(new_xlist)
            xmax = max(new_xlist)
            ymin = min(new_ylist)
            ymax = max(new_ylist)
            # print(xmin)
            # print(xmax)
            # print(ymin)
            # print(ymax)
            # print(xList_singlevalue)
            
            axis.hist2d(xList_singlevalue, yList_singlevalue, bins=100, range = [[xmin, xmax], [ymin, ymax]], norm = colors.LogNorm())

        axis.legend( handles = legendList , loc=1 )
        axis.set_xlim(90, 420)
        axis.set_ylim(-370, -40)

        #draw separated animals in separated graphs
        for animal in self.getAnimalList():
            axis = axes[self.getAnimalList().index(animal)+1]

            legendList=[]

            # print ("Compute trajectory of animal " + animal.name )
            xList2, yList2 = animal.getTrajectoryData( maskingEventTimeLine )
            
            # print ("Draw HeatMap of animal " + animal.name )
            
            # if scatter == True:
            #     axis.scatter( xList, yList, color= animal.getColor() , s=1 , linewidth=1, alpha=0.05, label= animal.RFID )
            #     legendList.append( mpatches.Patch(color=animal.getColor(), label=animal.RFID) )
            # else:
            #     axis.hist2d( xList, yList)
            #     # axis.hist2d( xList, yList, color= animal.getColor() , linestyle='-', linewidth=1, alpha=0.5, label= animal.RFID )
            
            xList_singlevalue = [item[0] for item in xList2]
            yList_singlevalue = [item[0] for item in yList2]
            
            new_xlist = [item for item in xList_singlevalue if not(math.isnan(item)) == True]
            new_ylist = [item for item in yList_singlevalue if not(math.isnan(item)) == True]
            
            xmin = min(new_xlist)
            xmax = max(new_xlist)
            ymin = min(new_ylist)
            ymax = max(new_ylist)
            # print(xmin)
            # print(xmax)
            # print(ymin)
            # print(ymax)
            # print(xList_singlevalue)
            
            axis.hist2d(xList_singlevalue, yList_singlevalue, bins=100, range = [[xmin, xmax], [ymin, ymax]], norm = colors.LogNorm())

        
        axis.legend( handles = legendList , loc=1 )

        fig.suptitle( title )

        if saveFile !=None:
            # print("Saving figure : " + saveFile )
            fig.savefig( saveFile, dpi=200)

        if show:
            plt.show()
        plt.close()
        

    def showMask(self, t ):
        '''
        show the mask of all animals in a figure
        '''

        fig, ax = plt.subplots()
        ax.set_xlim(90, 420)
        ax.set_ylim(-370, -40)

        for animal in self.getAnimalList():
            mask = animal.getBinaryDetectionMask( t )
            mask.showMask( ax=ax )

        plt.show()

        
    def getParticleDictionnary(self , start, end ):
        '''
        return the number of particle per frame
        '''

        query = "SELECT FRAMENUMBER, NUMPARTICLE FROM FRAME WHERE FRAMENUMBER>={} AND FRAMENUMBER<={}".format( start, end )

        # print ( "SQL Query: " + query )

        cursor = self.conn.cursor()
        cursor.execute( query )
        rows = cursor.fetchall()
        cursor.close()

        particleDictionnary = {}

        for row in rows:

            particleDictionnary[ row[0] ] = row[1]

        return particleDictionnary

    def getDetectionTable(self):
        """
        Returns detections as pandas table for all animals in pool.
        * adds location also in cm
        * adds time column in pandas timedelta
        * adds column to indicate detections in center region
        Returns:
            pd.DataFrame: detections as pandas table
        """

        data = defaultdict(list)
        for animal in self.getAnimalList():
            for frame, detection in animal.detectionDictionnary.items():
                data["RFID"]         .append(f"{animal.name}_{animal.RFID}")
                data["name"]         .append(f"{animal.name}")
                data["genotype"]     .append(f"{animal.genotype}")
                data['frame']        .append(frame)
                data['sec']          .append(frame / oneSecond)
                data['x']            .append(detection.massX)
                data['y']            .append(detection.massY)
                data['z']            .append(detection.massZ)

        df = pd.DataFrame(data)

        df["x_cm"] = (df.x - cornerCoordinates50x50Area[0][0]) / (cornerCoordinates50x50Area[1][0] - cornerCoordinates50x50Area[0][0]) * ARENA_SIZE
        df["y_cm"] = (df.y - cornerCoordinates50x50Area[1][1]) / (cornerCoordinates50x50Area[2][1] - cornerCoordinates50x50Area[1][1]) * ARENA_SIZE

        df[f"in_arena_center"] = (df["x_cm"] > CENTER_MARGIN) & \
                                 (df["y_cm"] > CENTER_MARGIN) & \
                                 (df["x_cm"] < (ARENA_SIZE - CENTER_MARGIN)) & \
                                 (df["y_cm"] < (ARENA_SIZE - CENTER_MARGIN))

        df.insert(3, "time", pd.to_timedelta(df.sec, unit="s"))

        return df.sort_values("time").reset_index(drop=True)

    def getSingleEventTable(self, event_name):
        """
        Returns pandas table containing all found events for given event name'
            * Primary mouse' meta data
            * Event start time
            * Event end time
            * Event duration
        Args:
            event_name (str): Event name e. g. Rearing
        Returns:
            DataFrame
        """
        data = defaultdict(list)
        for animal in self.getAnimalList():
            with mute_prints():
                eventTimeLine = EventTimeLine(self.conn,
                                                event_name,
                                                idA=animal.baseId,
                                                minFrame=self.detectionStartFrame,
                                                maxFrame=self.detectionEndFrame)

            for e in eventTimeLine.getEventList():
                data["RFID"]         .append(f"{animal.name}_{animal.RFID}")
                data["name"]         .append(f"{animal.name}")
                data["genotype"]     .append(f"{animal.genotype}")
                data["event_name"]   .append(event_name)
                data["start_sec"]    .append(e.startFrame / oneSecond)
                data["end_sec"]      .append(e.endFrame   / oneSecond)
                data["duration"]     .append(e.duration() / oneSecond)

        df = pd.DataFrame(data)
        df.insert(2, "time", pd.to_timedelta(df["start_sec"], unit="s"))

        return df.sort_values("time").reset_index(drop=True)

    def getAllEventsTable(self):
        """
        Returns pandas table containing all found events for all found event types'
            * Primary mouse' meta data
            * Events' start time
            * Events' end time
            * Events' duration
        Returns:
            DataFrame
        """
        with mute_prints():
            all_event_names = getAllEvents(connection=self.conn)

        event_table =  pd.concat([self.getSingleEventTable(event_name)
                                    for event_name in all_event_names]
                                , axis=0)
        return event_table.sort_values("time").reset_index(drop=True)

    def getNightStartStop(self):
        '''
        returns the Start and Stop frames of Night in the database.
        '''

        query = "SELECT STARTFRAME, ENDFRAME FROM EVENT WHERE NAME = 'night'"

        cursor = self.conn.cursor()
        cursor.execute(query)

        StartStopframes = cursor.fetchall()
        cursor.close()

        return StartStopframes


'''
Created on 6 sept. 2017

@author: Fab
'''
import sqlite3
from time import *
from lmtanalysis.Chronometer import Chronometer
from lmtanalysis.Animal import *
from lmtanalysis.Detection import *
from lmtanalysis.Measure import *
import numpy as np
from lmtanalysis.Event import *
from lmtanalysis.Measure import *
#from affine import Affine
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import sys
import matplotlib.pyplot as plt
from lmtanalysis.FileUtil import getFilesToProcess

def flush( connection ):
    ''' flush event in database '''
    deleteEventTimeLineInBase(connection, "onHouse" )

def reBuildEvent( connection, file, tmin=None, tmax=None, pool = None, showGraph = False ): 
    
    ''' use the pool provided or create it'''
    if ( pool == None ):
        pool = AnimalPool( )
        pool.loadAnimals( connection )
        pool.loadDetection( start = tmin, end = tmax )
    
    '''
    centerX = 512/2
    centerY = 424/2
    
    if ( showGraph ):
        plt.figure( 1 )
    '''
    
    # find threshold
    massZs = []
    headZs = []
    backZs = []

    '''
    for animal in pool.animalDictionnary.keys():
        
        animalA = pool.animalDictionnary[animal]        
        dicA = animalA.detectionDictionnary
        for t in dicA.keys():
            
            massZs.append( dicA[t].massZ )
            
            if dicA[t].frontZ > 0:
                headZs.append( dicA[t].frontZ )
            
            if dicA[t].backZ > 0:
                backZs.append( dicA[t].backZ )
            
    headZThreshold = np.percentile( headZs, 95 )
    massZThreshold = np.percentile( massZs, 95 )
    backZThreshold = np.percentile( backZs, 95 )
    '''
    headZThreshold = 50 -20
    massZThreshold = 148 -20
    backZThreshold = 30 -20 
    eventName = "onHouse"
    
    '''
    for animal in pool.animalDictionnary.keys():
        print(pool.animalDictionnary[animal])
        t = 335794
        animalA = pool.animalDictionnary[animal]        
        dicA = animalA.detectionDictionnary
        print( "mass: " , dicA[t].massZ )
        print( "head: " , dicA[t].frontZ )
        print( "back: " , dicA[t].backZ )
        
    
    quit()
    '''
    
    #deleteEventTimeLineInBase(connection, "onHouse" )
    
    for animal in pool.animalDictionnary.keys():
        # print(pool.animalDictionnary[animal])
        #
        # print ( "X is on the house")
        # print ( eventName )
                
        onHouseTimeLine = EventTimeLine( None, eventName , animal , None , None , None , loadEvent=False )
        centerZoneTimeLineDic = EventTimeLine( connection, "Center Zone" , animal ).getDictionary()
        
        result={}
        
        animalA = pool.animalDictionnary[animal]        
        dicA = animalA.detectionDictionnary
        
        for t in dicA.keys():
            
            if not t in centerZoneTimeLineDic:
                # avoid rearing on wall
                continue
            
            massOk = False
            headOk = False
            backOk = False
            
            if dicA[t].massZ > massZThreshold:
                massOk = True
            
            if dicA[t].frontZ > 0:
                if dicA[t].frontZ > headZThreshold:
                    headOk = True
                    
            if dicA[t].backZ > 0:
                if dicA[t].backZ > backZThreshold:
                    backOk = True
                            
            if massOk and headOk : # and backOk
                result[t] = True
                #print ( t , dicA[t].massZ , dicA[t].frontZ ) 
        
        
        '''
        print( np.percentile( allVals, 95 ) )
        print( np.percentile( allVals, 99 ) )

        plt.figure( 1 )
        plt.hist( allVals , bins = 1000 )
        plt.show()
        '''
                
        
        onHouseTimeLine.reBuildWithDictionnary( result )  
        onHouseTimeLine.removeEventsBelowLength( 5 )
        onHouseTimeLine.mergeCloseEvents( 30 )          
        onHouseTimeLine.endRebuildEventTimeLine(connection)
        # print( onHouseTimeLine )
        #quit()
        
    '''
    if ( showGraph ):
        plt.show()
    '''
        
    # log process
    from lmtanalysis.TaskLogger import TaskLogger
    t = TaskLogger( connection )
    t.addLog( "Build Event Wall Jump" , tmin=tmin, tmax=tmax )

     
    # print( "Rebuild event finished." )
        
if __name__ == '__main__':
    
    files = getFilesToProcess()

    chronoFullBatch = Chronometer("Full batch" )

    if ( files != None ):

        for file in files:
            
            # print ( "Processing file" , file )
            connection = sqlite3.connect( file )
            animalPool = AnimalPool( )
            animalPool.loadAnimals( connection )
            
            '''
            currentMinT = 335794-30
            currentMaxT = 335794+30
            '''
            
            
            currentMinT = 0
            #currentMaxT = 60*oneMinute
            
            #currentMaxT = 6*oneHour #3*oneDay
            currentMaxT = 3*oneDay
            
            animalPool.loadDetection( start = currentMinT, end = currentMaxT )
            reBuildEvent( connection, file, tmin=currentMinT, tmax=currentMaxT, pool = animalPool )
            
            
            
    chronoFullBatch.printTimeInS()
    print( "*** ALL JOBS DONE ***")

        
        
        
        
        
    


'''
Created on 6 sept. 2017

@author: Fab
'''
import sqlite3
from time import *
from lmtanalysis.Chronometer import Chronometer
from lmtanalysis.Animal import *
from lmtanalysis.Detection import *
from lmtanalysis.Measure import *
import matplotlib.pyplot as plt
import numpy as np
from lmtanalysis.Event import *
from lmtanalysis.Measure import *
from lmtanalysis.EventTimeLineCache import EventTimeLineCached

def flush( connection ):
    ''' flush event in database '''
    deleteEventTimeLineInBase(connection, "Passive oral-genital Contact" )

def reBuildEvent( connection, file, tmin=None, tmax=None , pool = None ): 

    if ( pool == None ):
        pool = AnimalPool( )
        pool.loadAnimals( connection )
        pool.loadDetection( start = tmin, end = tmax , lightLoad=True )

    ''' load the contact timeline for each pair of animals '''
    anogenitalSniffTimeLine = {}
    passiveAnogenitalSniffTimeLine = {}
    for animal in range( 1 , pool.getNbAnimals()+1 ):
        for idAnimalB in range( 1 , pool.getNbAnimals()+1 ):
            if ( animal == idAnimalB ):
                continue
            anogenitalSniffTimeLine[animal, idAnimalB] = EventTimeLineCached( connection, file, "Oral-genital Contact", animal, idAnimalB, minFrame=tmin, maxFrame=tmax )

            # create an empty timeline for other contacts
            passiveAnogenitalSniffTimeLine[idAnimalB, animal] = EventTimeLine(None, 'Passive oral-genital Contact', idAnimalB, animal, loadEvent=False)

    anogenitalSniffDico = {}
    for animal in range(1, pool.getNbAnimals() + 1):
        for idAnimalB in range(1, pool.getNbAnimals() + 1):
            if (animal == idAnimalB):
                continue
            anogenitalSniffDico[animal, idAnimalB] = anogenitalSniffTimeLine[animal, idAnimalB].getDictionary(minFrame=tmin, maxFrame=tmax)

            passiveAnogenitalSniffTimeLine[idAnimalB, animal].reBuildWithDictionnary(anogenitalSniffDico[animal, idAnimalB])
            passiveAnogenitalSniffTimeLine[idAnimalB, animal].endRebuildEventTimeLine(connection)


    # log process
    from lmtanalysis.TaskLogger import TaskLogger
    t = TaskLogger( connection )
    t.addLog( "Build Event Passive oral-genital Contact" , tmin=tmin, tmax=tmax )
                       
    # print( "Rebuild event finished." )
        
    


"""
Created on 17 January 2024

@author: P. Carrascosa
"""

"""
This code must be used with Jupyter Notebook or Jupyter Lab. It will allows you to change your genotypes in your
databases one by one.
The Goal is to then use an ipywidget tool, based on dataframes to compute Graphs and mixed model statistics.
"""

import sys
sys.path.insert(1, "../")
from lmtanalysis.FileUtil import getOneFileToProcess
import sqlite3
import os
import pandas as pd
from IPython.display import display, clear_output
from ipywidgets import interact_manual, interact, widgets, interactive_output, Output
import warnings
warnings.filterwarnings('ignore')

def connect_to_database():
    # Connect to the SQLite database
    file = getOneFileToProcess()
    conn = sqlite3.connect(file)
    return conn, file

def load_initial_data(conn):
    # Load data from the table into a DataFrame
    query = "SELECT * FROM ANIMAL;"
    df_sqlite = pd.read_sql_query(query, conn)
    return df_sqlite

def display_initial_table(df_sqlite):
    # Display the initial table
    df_display = display(df_sqlite, display_id='df_display')
    return df_display

def update_genotype(Selected_ID, Genotype, conn, df_display, file):
    try:
        # Start an SQLite transaction
        conn.execute("BEGIN;")

        # Update the database
        update_query = f"UPDATE ANIMAL SET GENOTYPE = ? WHERE ID = ?;"
        conn.execute(update_query, (Genotype, Selected_ID))

        # Commit the transaction
        conn.execute("COMMIT;")

        # Extract only the file name from the full path
        file_name = os.path.basename(file)
        print(f"Change applied to SQLite database {file_name} !")

        # Reload the DataFrame after modifications
        query = "SELECT * FROM ANIMAL;"
        df_sqlite = pd.read_sql_query(query, conn)

        # Clear the current display content
        clear_output(wait=True)

        # Display the updated table in the same display
        df_display.update(df_sqlite)

    except Exception as e:
        # Rollback in case of an error
        conn.execute("ROLLBACK;")
        print("Error:", str(e))

def close_connection(conn, output_widget):
    # Close the SQLite database connection
    conn.close()

    # Display the message in the specified Output widget
    with output_widget:
        clear_output(wait=True)
        print("Connection to the database closed! If you want to change another genotype in the same database, launch the cell again!")

def ChangeGenotypes():
    # Initial setup
    conn, file = connect_to_database()
    df_sqlite = load_initial_data(conn)
    df_display = display_initial_table(df_sqlite)

    # Create an Output widget for displaying messages
    output_widget = Output()

    # Function to update the value in the specified column
    @interact_manual(Selected_ID=df_sqlite['ID'], Genotype=widgets.Text(value=df_sqlite['GENOTYPE'].iloc[0]))
    def interact_update_genotype(Selected_ID, Genotype):
        update_genotype(Selected_ID, Genotype, conn, df_display, file)

    # Button to close the connection
    close_button = widgets.Button(description="Close Connection")
    close_button.on_click(lambda x: close_connection(conn, output_widget))

    # Display the widgets
    display(close_button, output_widget)

# if __name__ == '__main__':
#     ChangeGenotypes()

# # Run the main function
# ChangeGenotypes()


'''
Created on 6 sept. 2017

@author: Fab
'''
import sqlite3
from time import *

from affine import Affine

from lmtanalysis.Chronometer import Chronometer
from lmtanalysis.Animal import *
from lmtanalysis.Detection import *
from lmtanalysis.Measure import  SPEED_THRESHOLD_LOW
import numpy as np
from lmtanalysis.Event import *
from lmtanalysis.Measure import *

import matplotlib.pyplot as plt
import matplotlib.lines as mlines
from lmtanalysis.EventTimeLineCache import EventTimeLineCached
import copy
from copy import deepcopy

#eventName = "FollowZone New4"
eventName = "FollowZone"

FOLLOW_CORRIDOR_DURATION = 30 # in frames
FOLLOW_MAX_ANGLE = math.pi/4 # in radians
FOLLOW_DISTANCE_MAX_PIX = 2.5/scaleFactor # numeric value in cm to obtain pixels #2.5
FOLLOW_SPEED_MULTIPLICATOR_THRESHOLD = 2
FOLLOW_REMOVE_EVENT_BELOW_LEN = 7 # in frames        
FOLLOW_MERGE_EVENT_LEN_CRITERIA = 10 # in frames
            
def flush( connection ):
    ''' flush event in database '''
    # print("Flushing " , eventName )
    deleteEventTimeLineInBase(connection, eventName )


def isSameWay( detA, detB ):
    
    vectAX = detA.frontX - detA.backX
    vectAY = detA.frontY - detA.backY

    vectBX = detB.frontX - detB.backX
    vectBY = detB.frontY - detB.backY
    
    scalarProduct = vectAX * vectBX + vectAY * vectBY
    
    if ( scalarProduct >= 0 ): #same direction
        return True
    
    return False

def isAFollowingB( t , dicA , dicB ):
    
    # False is detection A does not exist at t
    if t not in dicA:
        return False
    
    # False is detection B does not exist at t
    if t not in dicB:
        return False
    
    # False if at least for the current t animals are not in the same direction. Could be wrong if animals are spinning around while following but filters out a number of complex situation
    if not isSameWay( dicA[t], dicB[t] ):
        return False
        
    detectionA = dicA[t]
    
    # check if in the past B was in A location
    for timeB in range( t-FOLLOW_CORRIDOR_DURATION,t+1 ):
        
        # check if detection B exists:
        if not timeB in dicB:
            continue
        
        detectionB = dicB[timeB]
        
        distance = detectionA.getDistanceTo( detectionB )  

        # discard invalid distances
        if distance == None:
            continue
                
        # discard if distance is too large between detections
        if distance > FOLLOW_DISTANCE_MAX_PIX:

            continue
                         
        # discard if animals are not going the same way ( scalar test )
        if not isSameWay( detectionA, detectionB ):
            continue
        
        # discard if angle between animals is not ok
        angleA = detectionA.getDirection()
        angleB = detectionB.getDirection()
        dbDif= math.atan2( math.sin(angleB-angleA), math.cos(angleB-angleA) )
        dif = math.fabs( dbDif )
        if ( dif > FOLLOW_MAX_ANGLE ):
            continue
        
        
        return True
        
    
    
    
    return False
    

def reBuildEvent( connection, file, tmin=None, tmax=None, pool = None ): 
    
    '''
    Event FollowZone:
    - ok: the two animals are moving at a speed >5 cm/s (SPEED_THRESHOLD_LOW)
    - ok: animals must not be in contact
    
    - ok: the angles between the two animals are less than 45 degrees apart considering head tail direction
    
    '''
    
    
    # create dedicated pool as we will alter detection pool with filters (I.e: we don't use the pool cache)
    
    pool = AnimalPool( )
    pool.loadAnimals( connection )
    pool.loadDetection( start = tmin, end = tmax )
    
    # filter detection by speed. Keep only the detection that are over SPEED_THRESHOLD_LOW.
    #pool.filterDetectionByInstantSpeed( pool.getAnimalList()[0].parameters.SPEED_THRESHOLD_LOW , 1000 ) # for the rat compatible version
    #pool.filterDetectionByInstantSpeed( SPEED_THRESHOLD_LOW*2 , 1000 )
    pool.filterDetectionByInstantSpeed( SPEED_THRESHOLD_LOW*FOLLOW_SPEED_MULTIPLICATOR_THRESHOLD , 1000 )
    
    # remove all detection where head and tail are missing
    pool.filterDetectionToKeepOnlyHeadTailDetection()
    
    
    # load the contact matrix dictionary    
    contactDic = {}    
    oralGenitalDic = {}
    for idAnimalA in pool.animalDictionnary:
        # print(pool.animalDictionnary[idAnimalA])
        for idAnimalB in pool.animalDictionnary:
            if( idAnimalA == idAnimalB ):
                continue
            contactDic[idAnimalA, idAnimalB] = EventTimeLineCached( connection, file, "Contact", idAnimalA, idAnimalB, minFrame=tmin, maxFrame=tmax ).getDictionary()    
            oralGenitalDic[idAnimalA, idAnimalB] = EventTimeLineCached( connection, file, "Oral-genital Contact", idAnimalA, idAnimalB, minFrame=tmin, maxFrame=tmax ).getDictionary()
            
    #init empty result EventTimeLine matrix
    followIsolatedTimeLine = {}
    for idAnimalA in pool.animalDictionnary:
        for idAnimalB in pool.animalDictionnary:
            if( idAnimalA == idAnimalB ):
                continue                                        
            followIsolatedTimeLine[idAnimalA, idAnimalB] = EventTimeLine( None, eventName , idAnimalA , idAnimalB , None , None , loadEvent=False )

    # init empty result timeline Dictionary.
    followIsolatedTimeLineDic = {}
    
    for idAnimalA in pool.animalDictionnary:

        animalA = pool.animalDictionnary[idAnimalA]
        dicA = animalA.detectionDictionnary
        
        for idAnimalB in pool.animalDictionnary:
            
            # discard if animals are the same.
            if( idAnimalA == idAnimalB ):
                continue
            
            # result dictionary for the current pair tested.            
            resultDic = {}

            animalB = pool.animalDictionnary[idAnimalB]
            dicB = animalB.detectionDictionnary
            
            # Starting "is A following B ?"
                        
            for t in dicB:            
                if isAFollowingB( t , dicA , dicB ):
                    resultDic[t]=True

            # remove all contact situation
            for t in contactDic[idAnimalA, idAnimalB]:
                if t in resultDic:
                    resultDic.pop( t )
            
            for t in oralGenitalDic[idAnimalA, idAnimalB]:
                if t in resultDic:
                    resultDic.pop( t )
            
            for t in oralGenitalDic[idAnimalB, idAnimalA]:
                if t in resultDic:
                    resultDic.pop( t )
            
            followIsolatedTimeLineDic[idAnimalA, idAnimalB] = resultDic

            
    #rebuild timelines with dictionaries
    for idAnimalA in pool.animalDictionnary:
        for idAnimalB in pool.animalDictionnary:
            if( idAnimalB == idAnimalA ):
                continue
            followIsolatedTimeLine[idAnimalA, idAnimalB].reBuildWithDictionnary( followIsolatedTimeLineDic[idAnimalA, idAnimalB] )
                        
            # filter out accidental events
            followIsolatedTimeLine[idAnimalA, idAnimalB].removeEventsBelowLength( FOLLOW_REMOVE_EVENT_BELOW_LEN )
            # create continuity in the events
            followIsolatedTimeLine[idAnimalA, idAnimalB].mergeCloseEvents( FOLLOW_MERGE_EVENT_LEN_CRITERIA )
                        
            followIsolatedTimeLine[idAnimalA, idAnimalB].endRebuildEventTimeLine(connection)
                
    # log process
    from lmtanalysis.TaskLogger import TaskLogger
    t = TaskLogger( connection )
    t.addLog( eventName , tmin=tmin, tmax=tmax )
              
    # print( "Rebuild event finished." )
    return
    
            
    
            
        
        
    


'''
Created on 6 sept. 2017

@author: Fab
'''
import sqlite3
from time import *
from lmtanalysis.Chronometer import Chronometer
from lmtanalysis.Animal import *
from lmtanalysis.Detection import *
from lmtanalysis.Measure import *
import numpy as np
from lmtanalysis.Event import *
from lmtanalysis.Measure import *
#from affine import Affine
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
from lmtanalysis.EventTimeLineCache import EventTimeLineCached

def flush( connection ):
    ''' flush event in database '''
    deleteEventTimeLineInBase(connection, "seq oral oral - oral genital" )
    deleteEventTimeLineInBase(connection, "seq oral geni - oral oral" )

def reBuildEvent( connection, file,  tmin=None, tmax=None, pool = None ): 
    
    ''' use the pool provided or create it'''
    if ( pool == None ):
        pool = AnimalPool( )
        pool.loadAnimals( connection )
        pool.loadDetection( start = tmin, end = tmax )
    '''
    Event OralSide Sequence
    '''
    nbAnimal = pool.getNbAnimals()
    
    contact = {}
    oralOral = {}
    oralGenital = {}
    oralOralDico = {}
    oralGenitalDico = {}
    
    EVENT_MIN_LEN = 10
    
    for idAnimalB in pool.animalDictionnary.keys():
        # print(pool.animalDictionnary[idAnimalB])
        meanSizeB = pool.animalDictionnary[idAnimalB].getMeanBodyLength( tmax = tmax )
        
        for animal in pool.animalDictionnary.keys():
            if( idAnimalB == animal ):
                continue
            
            '''
            contact[animal, idAnimalB] = EventTimeLine( connection, "Contact", animal, idAnimalB, minFrame=tmin, maxFrame=tmax )
            '''
            oralOral[animal, idAnimalB] = EventTimeLineCached( connection, file, "Oral-oral Contact", animal, idAnimalB, minFrame=tmin, maxFrame=tmax )
            oralGenital[animal, idAnimalB] = EventTimeLineCached( connection, file, "Oral-genital Contact", animal, idAnimalB, minFrame=tmin, maxFrame=tmax )
            
            oralOralDico[animal, idAnimalB] = oralOral[animal, idAnimalB].getDictionnary()
            oralGenitalDico[animal, idAnimalB] = oralGenital[animal, idAnimalB].getDictionnary()
        
    
    

    window = 60

    '''    
    oral oral followed by a oral-genital
    '''
    eventName ="seq oral oral - oral genital"
    # print ( eventName )
    
    for animal in range( 1 , nbAnimal+1 ):
        
        for idAnimalB in range( 1 , nbAnimal+1 ):
            if( animal == idAnimalB ):
                continue
        
                    
            selOO_OG_TimeLine = EventTimeLine( None, eventName , animal , idAnimalB , None , None , loadEvent=False )
            
            # print( selOO_OG_TimeLine.eventNameWithId )
                    
            oralOralTimeLine = oralOral[ animal , idAnimalB ]
            
            oralGeniDico = oralGenitalDico[ animal , idAnimalB ]         
            oralGeniTimeLine = oralGenital[ animal , idAnimalB ]
            
            for event in oralOralTimeLine.eventList:
                
                endFrame = event.endFrame
                
                if ( event.duration() <= EVENT_MIN_LEN ):  # discard too short events
                    continue

                for t in range ( endFrame , endFrame+window+1 ):
                    
                    if ( t in oralGeniDico ):
                        
                        event2 = oralGeniTimeLine.getEventAt( t )
                        
                        if ( event2.duration() >= EVENT_MIN_LEN ):
                        
                            start = event.startFrame
                            end = event2.endFrame
                            
                            selOO_OG_TimeLine.addEvent( Event( start, end ) )                        
                            break
                
            selOO_OG_TimeLine.endRebuildEventTimeLine(connection)
                    
    
    window = 60

    '''    
    oral A genital B followed by oral oral
    '''
    eventName ="seq oral geni - oral oral"
    # print ( eventName )
    
    for animal in range( 1 , nbAnimal+1 ):
        
        for idAnimalB in range( 1 , nbAnimal+1 ):
            if( animal == idAnimalB ):
                continue
        
                    
            selOG_OO_TimeLine = EventTimeLine( None, eventName , animal , idAnimalB , None , None , loadEvent=False )
            
            # print( selOG_OO_TimeLine.eventNameWithId )
                    
            oralGenitalTimeLine = oralGenital[ animal , idAnimalB ]
            
            ooDico = oralOralDico[ animal , idAnimalB ]         
            oralOralTimeLine = oralOral[ animal , idAnimalB ]
            
            for event in oralGenitalTimeLine.eventList:
                
                endFrame = event.endFrame
                
                if ( event.duration() <= EVENT_MIN_LEN ): # discard too short events
                    continue
                                
                for t in range ( endFrame , endFrame+window+1 ):
                    
                    if ( t in ooDico ):
                
                        event2 = oralOralTimeLine.getEventAt( t )
                                
                        if ( event2.duration() >= EVENT_MIN_LEN ):
                                
                            start = event.startFrame
                            end = event2.endFrame
                            
                            selOG_OO_TimeLine.addEvent( Event( start, end ) )                        
                            break
                
            selOG_OO_TimeLine.endRebuildEventTimeLine(connection)
                    
    
        
    # log process
    from lmtanalysis.TaskLogger import TaskLogger
    t = TaskLogger( connection )
    t.addLog( "Build Event Oral Side Sequence" , tmin=tmin, tmax=tmax )

                    
    # print( "Rebuild event finished." )
        
    


'''
Copyright (C)
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program. If not, see https://tldrlegal.com/license/gnu-general-public-license-v3-(gpl-3)#fulltext. 

For license issues, please contact:

Paul Carrascosa
Institut de Genomique Fonctionnelle
141 Rue de la Cardonille
34000, Montpellier

Email: paul.carrascosa@igf.cnrs.fr or damien.huzard@igf.cnrs.fr
'''

from setuptools import setup, find_packages
import pathlib
import os

path=pathlib.Path(__file__).parent.resolve()

try:
    with open(os.path.join(path,'README.md'),encoding='utf-8') as readme:
        long_description=readme.read()
except Exception:
    long_description=''


setup(name='LWTools',version='1.1.4',author='Paul Carrascosa, Damien Huzard',
    author_email='paul.carrascosa@igf.cnrs.fr',
    description='Tool for LMT data analysis',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/PaulCarrascosa/LMT_Widget_Tool-LWT',
    packages=find_packages(),
    include_package_data=True,
    package_data={
        'LWTools':['*.py''*.ipynb','*.txt','*.png','*.pb','*.jpg','*.csv','*.index','*.data-00000-of-00001']
    },
    license='GNU General Public License v3 (GPLv3)',
    classifiers=[
    'Programming Language :: Python :: 3.10',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Operating System :: OS Independent',
    ],
    keywords='Live Mouse Tracker, LMT, Tool, Data analysis',
    python_requires='>=3.10',
    install_requires=[
        'setuptools>=61.0',
        'tqdm==4.65.0',
    	'matplotlib==3.5.2',
    	'lxml==4.9.1',
    	'pandas==1.5.0',
    	'affine==2.3.1',
    	'networkx==2.8.5',
    	'seaborn==0.11.2',
    	'dabest==2023.2.14',
    	'statsmodels==0.13.2',
    	'tabulate==0.8.10',
    	'ipywidgets==8.0.3',
    	'jupyterlab==3.5.0',
        'scipy==1.9.3',
        'numpy==1.22.3',
        
    ]   
)



'''
Created on 6 sept. 2017

@author: Fab
'''
import sqlite3
from time import *
from lmtanalysis.Chronometer import Chronometer
from lmtanalysis.Animal import *
from lmtanalysis.Detection import *
from lmtanalysis.Measure import *
import matplotlib.pyplot as plt
import numpy as np
from lmtanalysis.Event import *
from lmtanalysis.Measure import *

def flush( connection ):
    ''' flush event in database '''
    deleteEventTimeLineInBase(connection, "Huddling" )

def reBuildEvent( connection, file, tmin=None, tmax=None, pool = None , showGraph = False ): 
    
    ''' use the pool provided or create it'''
    if ( pool == None ):
        pool = AnimalPool( )
        pool.loadAnimals( connection )
        #pool.loadDetection( start = tmin, end = tmax )
    
    
    for idAnimalA in pool.animalDictionnary:
        
        threshold = 0.5
        animal = pool.animalDictionnary[idAnimalA]
        # print("Computing huddling for animal " , animal )
                
        huddlingTimeLine = EventTimeLine( connection, "Huddling", idAnimalA, minFrame=tmin, maxFrame=tmax, loadEvent=False )

        result = {}
        
        for t in range( tmin, tmax+1 ):
            # if t%10000 == 0:
            #     print ( "current t", t )
                            
            mask = animal.getBinaryDetectionMask( t )
            if mask == None:
                continue
            roundness = mask.getRoundness()
            if roundness == None:
                continue
            #print ( idAnimalA , t , roundness )
            #if roundness > 1-threshold and roundness < 1+threshold:
            if roundness < 1.85: # and roundness > 1:
                result[t] = True  
            
        huddlingTimeLine.reBuildWithDictionnary( result )
        huddlingTimeLine.endRebuildEventTimeLine(connection)
    
        
    # log process
    from lmtanalysis.TaskLogger import TaskLogger
    t = TaskLogger( connection )
    t.addLog( "Build Event Huddling" , tmin=tmin, tmax=tmax )

               
    # print( "Rebuild event finished." )
        
    


'''
Created on 6 sept. 2017

@author: Fab
'''
import sqlite3
from time import *
from lmtanalysis.Chronometer import Chronometer
from lmtanalysis.Animal import *
from lmtanalysis.Detection import *
from lmtanalysis.Measure import *
import matplotlib.pyplot as plt
import numpy as np
from lmtanalysis.Event import *
from lmtanalysis.Measure import *
from lmtanalysis.EventTimeLineCache import EventTimeLineCached
import networkx as nx

def flush( connection ):
    ''' flush event in database '''
    deleteEventTimeLineInBase(connection, "Nest4_" )
    '''
    could extends to those:
    deleteEventTimeLineInBase(connection, "Nest3" )
    deleteEventTimeLineInBase(connection, "Group2" )
    deleteEventTimeLineInBase(connection, "Group3" )
    deleteEventTimeLineInBase(connection, "Group4" )
    '''


def reBuildEvent( connection, file, tmin=None, tmax=None , pool = None ):
    '''
    Nest 3
    Nest 4
    Group 2
    Group 3
    Group 4
    ''' 
    # print("[NEST 4] : Assume that there is no occlusion")
    
    if ( pool == None ):
        pool = AnimalPool( )
        pool.loadAnimals( connection )
        pool.loadDetection( start = tmin, end = tmax , lightLoad=True )
    
    # check if given max is more than available detection 
    '''
    maxT = pool.getMaxDetectionT()
    if ( tmax > maxT ):
        tmax = maxT
    '''
    
    #pool.loadDetection( start = tmin, end = tmax )
    
    if ( len ( pool.getAnimalList() ) != 4 ):
        # print( "[NEST4 Cancelled] 4 animals are required to build nest 4.")
        return
    
    contact = {}
    
    
    for animal in range( 1 , 5 ):
        for idAnimalB in range( 1 , 5 ):
            if animal != idAnimalB:    
                contact[animal,idAnimalB] = EventTimeLineCached( connection, file, "Contact", animal, idAnimalB, minFrame=tmin, maxFrame=tmax ).getDictionnary() #fait une matrice de tous les contacts à deux possibles
    
    stopDictionnary = {}
        
    for animal in range( 1 , 5 ):
        stopDictionnary[animal] = EventTimeLineCached( connection, file, "Stop", animal, minFrame=tmin, maxFrame=tmax ).getDictionnary()
    
    
    
    '''
    nest3TimeLine = {}
    
    for animal in range( 1 , 5 ):
        nest3TimeLine = EventTimeLine( None, "Nest3" , animal, loadEvent=False )
    '''
    nest4TimeLine = EventTimeLine( None, "Nest4_" , loadEvent=False )
    
    pool.loadAnonymousDetection()
    
    '''
    group2TimeLine = {}
    for animal in range( 1 , 5 ):
        for idAnimalB in range( 1 , 5 ):
            if ( animal != idAnimalB ):
                group2TimeLine[animal,idAnimalB] = EventTimeLine( None, "Group2" , animal , idAnimalB , loadEvent=False )

    group3TimeLine = {}
    for animal in range( 1 , 5 ):
        for idAnimalB in range( 1 , 5 ):
            if( animal != idAnimalB ):
                for idAnimalC in range( 1 , 5 ):
                    if ( animal != idAnimalC and idAnimalB != idAnimalC ):
                        group3TimeLine[animal,idAnimalB] = EventTimeLine( None, "Group3" , animal , idAnimalB , idAnimalC, loadEvent=False )
    
    group4TimeLine = EventTimeLine( None, "Group4" , loadEvent=False )
    '''
        
    
    animalList = pool.getAnimalList() 
    
    result = {}
    
    for t in range( tmin, tmax+1 ):
        
        
        isNest = False
        
        nbAnimalAtT = 0
        animalDetectedList = []
        
        
        anonymousDetectionList = pool.getAnonymousDetection( t )
        
        for animal in animalList:
            if t in animal.detectionDictionnary:                
                animalDetectedList.append( animal )
        
        #print( str(t) + " : " + str( nbAnimalAtT ) )
                    
    
        #print("TEST")
        graph = nx.Graph()
        # add nodes
        
        for animal in animalDetectedList:
            graph.add_node( animal )
            nbAnimalAtT+=1
            
        for animalA in animalDetectedList:
            for animalB in animalDetectedList:
                if animalA != animalB:
                    # add an edge
                    if t in contact[animalA.baseId,animalB.baseId]:
                        graph.add_edge( animalA, animalB )
        
        # check with anonymous detection. Check contact
        if anonymousDetectionList!= None:
            
            nbAnimalAtT+=len(anonymousDetectionList)
            
            # manage anonymous
            #print( t , "manage anonymous")
            '''
            # load all masks
            for animal in animalDetectedList:
                animal.loadMask( t )
            '''
            
            for detectionA in anonymousDetectionList: # anonymous with anonymous
                for detectionB in anonymousDetectionList: # anonymous with anonymous
                    if detectionA != detectionB:
                        distance = detectionA.getDistanceTo( detectionB )
                        if distance != None:
                            if distance < DISTANCE_CONTACT_MASS_CENTER:
                                graph.add_edge( detectionA, detectionB )
                                #print("Adding edge with mask (det anonymous to det anonymous)")
                    
            for detection in anonymousDetectionList:
                for animal in animalDetectedList:
                    distance = detection.getDistanceTo(animal.getDetectionAt( t ) )
                    if distance != None:
                        if distance < DISTANCE_CONTACT_MASS_CENTER:
                            #if detection.getMask().isInContactWithMask( animal.getDetectionAt ( t ).getMask() ):
                            graph.add_edge( animal, detection )
                            #print("Adding edge with mask")
        
        # list of CC from the biggest to the smallest
        #listCC = sorted(nx.connected_components( graph ), key=len, reverse=True)
        
        if nbAnimalAtT == 0:            
            isNest = True
        
        # list of CC from the biggest to the smallest
        listCC = sorted(nx.connected_components( graph ), key=len, reverse=True)
        
        #largestCC = len ( max(nx.connected_components( graph ), key=len) )

        '''
        for animal in animalList:
            if t in animal.detectionDictionnary:
                nbAnimalAtT+=1
                animalDetectedList.append( animal )
        
        #print( str(t) + " : " + str( nbAnimalAtT ) )
        
        if nbAnimalAtT == 0:            
            isNest = True
            
        if not isNest:
            #print("TEST")
            graph = nx.Graph();
            # add nodes
            for animal in animalDetectedList:
                graph.add_node( animal )
            for animalA in animalDetectedList:
                for animalB in animalDetectedList:
                    if animalA != animalB:
                        # add an edge
                        if t in contact[animalA.baseId,animalB.baseId]:
                            graph.add_edge( animalA, animalB )
            
            
            
            
            # check connected components. If the biggest group gets all animal, we got a nest4
            largestCC = len ( max(nx.connected_components( graph ), key=len) )
            
            #print( str( t ) + " : " + str ( len( largestCC ) ) )
            
            #print( str( t ) + " : " + str ( largestCC ) + " / " + str( nbAnimalAtT ) )
        '''    
        if len ( listCC ) == 0 :
            continue
        
        #print( t , len ( listCC[0] ) , nbAnimalAtT )
        
        if len ( listCC[0] ) == nbAnimalAtT:
        
        #if largestCC == nbAnimalAtT :
            
            # check if animals in the nest are stopped.
            allStoppedInBiggestGroup = True
            for animal in animalDetectedList:
                if isinstance( animal , Animal ):
                    if not ( t in stopDictionnary[animal.baseId] ):
                        allStoppedInBiggestGroup = False
                        break

            if allStoppedInBiggestGroup:
                isNest= True                     
                     
        if isNest == True:
            #print( "ADD PUNCTUAL")
            result[t] = True;
            
            
            
    nest4TimeLine.reBuildWithDictionnary( result )
    # remove very small events
    nest4TimeLine.removeEventsBelowLength( 2 )
    # merge flashing events
    nest4TimeLine.mergeCloseEvents( 3 )
    nest4TimeLine.endRebuildEventTimeLine(connection)
        
    
    '''
    for animal in range( 1 , 5 ):
        
        for idAnimalB in range( 1 , 5 ):
            if( animal == idAnimalB ):
                continue
            
            for idAnimalC in range( 1 , 5 ):
                if( animal == idAnimalC ):
                    continue
                if( idAnimalB == idAnimalC ):
                    continue
                
                for idAnimalD in range( 1 , 5 ):
                    if( animal == idAnimalD ):
                        continue
                    if( idAnimalB == idAnimalD ):
                        continue
                    if( idAnimalC == idAnimalD ):
                        continue
                
                    eventName = "Group4"        
                    print ( eventName )
                    
                    groupTimeLine = EventTimeLine( None, eventName , animal , idAnimalB , idAnimalC , idAnimalD , loadEvent=False )
                    
                    result={}
                    
                    dicA = contact[ animal ].getDictionnary()
                    dicB = contact[ idAnimalB ].getDictionnary()
                    dicC = contact[ idAnimalC ].getDictionnary()
                    dicD = contact[ idAnimalD ].getDictionnary()
                    
                    dicGroup2A = group2[ animal ].getDictionnary()
                    dicGroup2B = group2[ idAnimalB ].getDictionnary()
                    dicGroup2C = group2[ idAnimalC ].getDictionnary()
                    dicGroup2D = group2[ idAnimalD ].getDictionnary()
                    
                    for t in dicA.keys():
                        if ( t in dicB and t in dicC and t in dicD ):
                            if ( t in dicGroup2A or t in dicGroup2B or t in dicGroup2C or t in dicGroup2D):
                                continue
                            else:
                                result[t]=True
                    
    groupTimeLine.reBuildWithDictionnary( result )
    
    groupTimeLine.endRebuildEventTimeLine(connection)
          
    '''                
        
    # log process
    from lmtanalysis.TaskLogger import TaskLogger
    t = TaskLogger( connection )
    t.addLog( "Build Event Nest4" , tmin=tmin, tmax=tmax )
          
    
    # print( "Rebuild event finished." )
        
    


'''
Created on 6 sept. 2017

@author: Fab
'''
import sqlite3
from time import *
from lmtanalysis.Chronometer import Chronometer
from lmtanalysis.Animal import *
from lmtanalysis.Detection import *
from lmtanalysis.Measure import *
import matplotlib.pyplot as plt
import numpy as np
from lmtanalysis.Event import *
from lmtanalysis.Measure import *
from lmtanalysis.EventTimeLineCache import EventTimeLineCached

def flush( connection ):
    ''' flush event in database '''
    deleteEventTimeLineInBase(connection, "Group3" )


def reBuildEvent( connection, file, tmin=None, tmax=None, pool = None  ):
    
    pool = AnimalPool( )
    pool.loadAnimals( connection )
    #pool.loadDetection( start = tmin, end = tmax )
    
    if pool.getNbAnimals() <= 2:
        # print( "Not enough animals to process group 3")
        return

    contact = {}
    for animal in range( 1 , 5 ):
        for idAnimalB in range( 1 , 5 ):
            if ( animal == idAnimalB ):
                continue
            contact[animal, idAnimalB] = EventTimeLineCached( connection, file, "Contact", animal, idAnimalB, minFrame=tmin, maxFrame=tmax )

    for animal in range( 1 , 5 ):
        
        for idAnimalB in range( 1 , 5 ):
            if( animal == idAnimalB ):
                continue
            
            for idAnimalC in range( 1 , 5 ):
                if( animal == idAnimalC ):
                    continue
                if( idAnimalB == idAnimalC ):
                    continue
                
                for idAnimalD in range( 1 , 5 ):
                    if( animal == idAnimalD ):
                        continue
                    if( idAnimalB == idAnimalD ):
                        continue
                    if( idAnimalC == idAnimalD ):
                        continue
                
                    eventName = "Group3"        
                    # print ( eventName )
                    
                    groupTimeLine = EventTimeLine( None, eventName , animal , idAnimalB , idAnimalC , None , loadEvent=False )
                    
                    result={}
                    
                    dicAB = contact[ animal , idAnimalB ].getDictionnary()
                    dicAC = contact[ animal , idAnimalC ].getDictionnary()
                    dicAD = contact[ animal , idAnimalD ].getDictionnary()
                    dicBC = contact[ idAnimalB , idAnimalC ].getDictionnary()
                    dicBD = contact[ idAnimalB , idAnimalD ].getDictionnary()
                    dicCD = contact[ idAnimalC , idAnimalD ].getDictionnary()
                    
                    for t in dicAB.keys():
                        if ( t in dicAC or t in dicBC ):
                            if ( t in dicAD or t in dicBD or t in dicCD ):
                                continue
                            else:
                                result[t]=True
                    
                    for t in dicAC.keys():   
                        if ( t in dicBC ):
                            if ( t in dicAD or t in dicBD or t in dicCD ):
                                continue
                            else:
                                result[t]=True
                    
                groupTimeLine.reBuildWithDictionnary( result )
                
                groupTimeLine.endRebuildEventTimeLine(connection)
                    
        
    # log process
    from lmtanalysis.TaskLogger import TaskLogger
    t = TaskLogger( connection )
    t.addLog( "Build Event Group 3" , tmin=tmin, tmax=tmax )
      
                    
    # print( "Rebuild event finished." )
        
    


'''
Created on 12 sept. 2017

@author: Fab
'''

import math
from lmtanalysis.Measure import *
import zlib
from lxml import etree

import matplotlib
#matplotlib fix for mac
#matplotlib.use('TkAgg')

import matplotlib.pyplot as plt

class Mask():
    '''
    Binary mask of an animal at a given t.
    '''    
    
    def __init__(self, data, color="black"  ):
        
        x = 0
        y = 0
        w = 0
        h = 0
        boolMaskData = None
        
        tree = etree.fromstring( data )
        for user in tree.xpath("/root/ROI/boundsX"):
            x = int(user.text)
        for user in tree.xpath("/root/ROI/boundsY"):
            y = int(user.text)
        for user in tree.xpath("/root/ROI/boundsW"):
            w = int(user.text)
        for user in tree.xpath("/root/ROI/boundsH"):
            h = int(user.text)
        for user in tree.xpath("/root/ROI/boolMaskData"):
            boolMaskData = user.text

        self.x = x
        self.y = y 
        self.width = w
        self.height = h        
        self.unzip( boolMaskData )
        self.color = color
        
        #mask = Mask( x , y , w , h , boolMaskData, self.getColor() )

    
    def getPerimeter(self ):
        
        dicA = {}

        def isOnPerimeter( x , y ):
            
            nonlocal dicA

            for xx in range( -1,2 ):
                for yy in range( -1,2 ):
                    if ( x+xx , y+ yy ) not in dicA:
                        return True                
            
        # create a mask map
        for i in range( len( self.pointsX ) ):
            dicA[(self.pointsX[i], self.pointsY[i])] = True
            
        nbPointPerimeter = 0
        
        for i in range( len ( self.pointsX ) ):
            
            x = self.pointsX[i]
            y = self.pointsY[i]
            
            if isOnPerimeter( x , y ):
                nbPointPerimeter+=1
            
        return nbPointPerimeter
        
        
        
    def getRoundness(self):
        
        '''
        if not hasattr(self, 'pointX'):
            print( "none")
            return None
        '''
        try:
            area = len( self.pointsX )
            '''
            circularity = 4.0 * area / ( self.getPerimeter() * longAxis );
            '''
            roundness = ( self.getPerimeter()**2 ) / ( 4.0 * math.pi * area )
            return roundness;
        except:
            print("Problem in Mask.getRoundness()" )
            
        return None
        
    def showMask(self , color = None , ax = None ):
        '''
        show the mask in a figure
        '''
        if ( color == None ):
            color = self.color
        
        if ( ax == None ):
            fig, ax = plt.subplots()
            ax.scatter( self.pointsX, self.pointsY , c=color )
            plt.show()
        else:
            ax.scatter( self.pointsX, self.pointsY , c=color )
    
    def isInContactWithMask(self, mask ):
        
        dicA = {}
        # create a first map with the self mask and 1 pix of dilation
        for i in range( len( self.pointsX ) ):
            for x in range( -1,2 ):
                for y in range( -1,2 ):
                    dicA[(self.pointsX[i]+x, self.pointsY[i]+y)] = True
        # check if a point overlaps
        for i in range( mask.pointsX ):
            if (mask.pointsX[i], mask.pointsY[i]) in dicA:
                return True
        
        return False
            
    def getNbPoint(self):
        return len( self.pointsX )
    
    def unzip(self, maskDataZipped ):
        
        # re fill 0 and put space instead of : separator
        if maskDataZipped == None:
            return    
        s = maskDataZipped.split(":")
        s2 = ""
        for value in s:
            if ( len(value) == 1 ):
                s2+="0"
            s2+=value+" "
            
        #print ( s2 )
        #print ("************")
        b = bytearray.fromhex( s2 )
        
        #print("uncompressed: ")
        uncompressed= zlib.decompress( b )
        
        #print ( uncompressed )
                    
        self.pointsX= []
        self.pointsY= []
                
        index = 0
        for y in range( self.y , self.y+ self.height ):
            for x in range( self.x , self.x+ self.width ):
                        
                if ( uncompressed[index] == 1 ):
                    self.pointsX.append( x )
                    self.pointsY.append( -y )
                
                index+=1
                
        
        


'''
Created on 29 ao�t 2018

@author: Fab
'''


import glob
from tkinter.dialog import Dialog
from tkinter.filedialog import askopenfilename, askdirectory
import tkinter as tk
import unittest
from random import randrange, random



'''
Provide a dialog to ask for either files or folder to process.
In case of folder, this use glob to recursively get all .sqlite files in sub folders.
'''
def getFilesToProcess():


    root = tk.Tk()
    root.withdraw()
    root.update()

    d = Dialog(
        title="Select input for processing", text="Select file(s) or folder for processing", bitmap = 'question',
        strings=('Files', 'Folder', 'Cancel'), default=0 )

    root.focus_force()
    files = None    
    if(  d.num == 0 ):
        files = askopenfilename( title="Choose a set of file to process", multiple=1, filetypes = (("sqlite files","*.sqlite"),("all files","*.*") )  )

    if ( d.num == 1):
        
        folder = askdirectory( title= "Choose a directory to process")
        print ("Folder: " , folder )
        folder = folder+"/**/*.sqlite"
        print( "Fetching files...")
        files = []
        for file in glob.glob( folder, recursive=True ):
            print( file , "found.")
            files.append( file )
    
    d.destroy()
    root.destroy()
    
    return files

def getFolderToProcess():
    root = tk.Tk()
    root.withdraw()
    root.update()

    d = Dialog(
        title="Select input for processing", text="Select folder for processing", bitmap='question',
        strings=('Folder', 'Cancel'), default=0)

    root.focus_force()
    folder = None
    if d.num == 0:
        folder = askdirectory(title="Choose a directory to process")

    d.destroy()
    root.destroy()

    return folder

# # Exemple d'utilisation :
# selected_folder = getFolderToProcess()
# # if selected_folder:
# #     print("Selected folder:", selected_folder)
# # else:
# #     print("Operation cancelled.")

def getJsonFileToProcess():
    root = tk.Tk()
    root.withdraw()
    root.update()

    d = Dialog(
        title="Select json file for processing", text="Select file for processing", bitmap='question',
        strings=('File', 'Cancel'), default=0)

    root.focus_force()
    file = None
    if (d.num == 0):
        file = askopenfilename(title="Choose a file to process", multiple=0,
                                filetypes=(("json files", "*.json"), ("all files", "*.*")))

    d.destroy()
    root.destroy()

    return file


def getJsonFilesToProcess():
    root = tk.Tk()
    root.withdraw()
    root.update()

    d = Dialog(
        title="Select input for processing", text="Select file(s) or folder for processing", bitmap='question',
        strings=('Files', 'Folder', 'Cancel'), default=0)

    root.focus_force()
    files = None
    if (d.num == 0):
        files = askopenfilename(title="Choose a set of file to process", multiple=1,
                                filetypes=(("json files", "*.json"), ("all files", "*.*")))

    if (d.num == 1):

        folder = askdirectory(title="Choose a directory to process")
        print("Folder: ", folder)
        folder = folder + "/**/*.json"
        print("Fetching files...")
        files = []
        for file in glob.glob(folder, recursive=True):
            print(file, "found.")
            files.append(file)

    d.destroy()
    root.destroy()

    return files

def getOneFileToProcess():
    root = tk.Tk()
    root.withdraw()
    root.update()

    d = Dialog(
        title="Select sqlite file for processing", text="Select sqlite file for processing", bitmap='question',
        strings=('File', 'Cancel'), default=0)

    root.focus_force()
    file = None
    if (d.num == 0):
        file = askopenfilename(title="Choose a sqlite file to process", multiple=0,
                                filetypes = (("sqlite files","*.sqlite"),("all files","*.*")))

    d.destroy()
    root.destroy()

    return file

def getCsvFileToProcess():
    root = tk.Tk()
    root.withdraw()
    root.update()

    d = Dialog(
        title="Select csv file for processing", text="Select csv file for processing", bitmap='question',
        strings=('File', 'Cancel'), default=0)

    root.focus_force()
    file = None
    if (d.num == 0):
        file = askopenfilename(title="Choose a csv file to process", multiple=0,
                                filetypes=(("csv files", "*.csv"), ("all files", "*.*")))

    d.destroy()
    root.destroy()

    return file

def getCsvFilesToProcess():
    root = tk.Tk()
    root.withdraw()
    root.update()

    d = Dialog(
        title="Select csv files for processing", text="Select file(s) or folder for processing", bitmap='question',
        strings=('Files', 'Folder', 'Cancel'), default=0)

    root.focus_force()
    files = None
    if (d.num == 0):
        files = askopenfilename(title="Choose a set of file to process", multiple=1,
                                filetypes=(("csv files", "*.csv"), ("all files", "*.*")))

    if (d.num == 1):

        folder = askdirectory(title="Choose a directory to process")
        print("Folder: ", folder)
        folder = folder + "/**/*.sqlite"
        print("Fetching files...")
        files = []
        for file in glob.glob(folder, recursive=True):
            print(file, "found.")
            files.append(file)

    d.destroy()
    root.destroy()

    return files

def addJitter(x, jit):
    newX = []
    for item in x:
        addedJitter = (random() * 2 - 1) * jit
        newX.append(item + addedJitter)

    return newX


def extractPValueFromLMMResult( result, keyword ):
    r = result.summary().as_text()
    for l in r.split("\n"):
        if keyword in l:
            print (l)
            lineWithoutSpace = ' '.join(l.split())
            pValue = float( lineWithoutSpace.split(" ")[4] )
            sign = 1
            print( "test: " , lineWithoutSpace.split(" ")[1] )
            if float( lineWithoutSpace.split(" ")[1] ) < 0:
                sign=-1
            print ( "P VALUE :" , pValue )
            print ( "SIGN :" , sign )
            return pValue, sign

def getFigureBehaviouralEventsLabelsFrench(event):
    behaviouralEventsLabels = {"Stop isolated": 'repos isole',
                         "Move isolated": 'mouvement isole',
                         "Break contact": 'rupture de contact',
                         "Get away": 'echappement',
                         "Social approach": 'approche',
                         # "Approach rear": 'approach reared mouse',
                         "Approach contact": 'approche avant contact',
                         "Contact": 'contact',
                         "Oral-oral Contact": 'nez-nez',
                         "Oral-genital Contact": 'nez-anogenital',
                         "Side by side Contact": 'cote-a-cote',
                         "Side by side Contact, opposite way": 'cote-a-cote, tete beche',
                         "seq oral oral - oral genital": 'nez-nez & nez-anogenital',
                         "seq oral geni - oral oral": 'nez-anogenital & nez-nez',
                         "FollowZone Isolated": 'poursuite',
                         "FollowZone": 'poursuite',
                         "Train2": 'train2'
                         # , "longChase": 'long chase'
                         }
    return behaviouralEventsLabels[event]

def getFigureBehaviouralEventsLabels(event):
    behaviouralEventsLabels = {
                        'totalDistance': 'distance',
                        "Stop isolated": 'single idle',
                         "Move isolated": 'single move',
                         "Move in contact": 'move in contact',
                         "WallJump": 'jumps',
                         "Rear isolated": 'rearing',
                         "Rear in contact": 'rearing in contact',
                         "Break contact": 'break contact',
                         "Get away": 'get away',
                         "Social approach": 'approach social range',
                         #"Approach rear": 'approach reared mouse',
                         "Approach contact": 'approach contact',
                         "Contact": 'contact',
                         "Group2": 'group of 2', 
                         "Group3": 'group of 3', 
                         "Oral-oral Contact": 'nose-nose',
                         "Oral-genital Contact": 'nose-anogenital',
                         "Side by side Contact": 'side-side',
                         "Side by side Contact, opposite way": 'side-side, head-to-tail',
                         "seq oral oral - oral genital": 'nose-nose & nose-anogenital',
                         "seq oral geni - oral oral": 'nose-anogenital & nose-nose',
                         "FollowZone Isolated": 'follow0',
                         "FollowZone": 'follow',
                         "Train2": 'train2',
                         "Group 3 make": 'make group 3',
                         "Group 4 make": 'make group 4',
                         "Break contact": 'break contact',
                         "Group 3 break": 'break group 3',
                         "Group 4 break": 'break group 4',
                         
                         'Oral-oral Contact exclusive': 'nose-nose (x)',
                           'Side by side Contact exclusive': 'side-side (x)',
                           'Oral-genital Contact exclusive': 'nose-anogenital (x)',
                           'Passive oral-genital Contact exclusive': 'passive nose-anogenital (x)',
                           'Side by side Contact, opposite way exclusive': 'side-side head-to-tail (x)',
                           'Oral-oral and Side by side Contact exclusive': 'nose-nose & side-side (x)',
                           'Oral-genital and Side by side Contact, opposite way exclusive': 'nose-anogenital & side-side head-to-tail (x)',
                           'Oral-genital passive and Side by side Contact, opposite way exclusive':
                               'passive nose-anogenital & side-side head-to-tail (x)', 'Other contact exclusive': 'other contacts (x)',
                           'Move isolated exclusive': 'single move (x)',
                           'Stop isolated exclusive': 'single idle (x)',
                           'Undetected exclusive': 'undetected (x)'
                         # , "longChase": 'long chase'
                         }
    
    return behaviouralEventsLabels[event]







behaviouralEventOneMouse = ["Move isolated", "Move in contact", "WallJump", "Stop isolated", "Rear isolated", "Rear in contact",
    "Contact", "Group2", "Group3", "Oral-oral Contact", "Oral-genital Contact", "Side by side Contact", "Side by side Contact, opposite way",
    "Train2", "FollowZone",
    "Social approach", "Approach contact",
    "Group 3 make", "Group 4 make", "Get away", "Break contact",
    "Group 3 break", "Group 4 break"
    ]

behaviouralEventOneMouseDic = {' TotalLen': ["Move isolated", "Move in contact", "WallJump", "Stop isolated", "Rear isolated", "Rear in contact",
    "Contact", "Group2", "Group3", "Oral-oral Contact", "Oral-genital Contact", "Side by side Contact", "Side by side Contact, opposite way",
    "Train2", "FollowZone"],
    
    ' Nb': ["Move isolated", "Move in contact", "WallJump", "Stop isolated", "Rear isolated", "Rear in contact",
    "Contact", "Group2", "Group3", "Oral-oral Contact", "Oral-genital Contact", "Side by side Contact", "Side by side Contact, opposite way",
    "Train2", "FollowZone",
    "Social approach", "Approach contact",
    "Group 3 make", "Group 4 make", "Break contact",
    "Group 3 break", "Group 4 break"],
    
    ' MeanDur': ["Move isolated", "Move in contact", "WallJump", "Stop isolated", "Rear isolated", "Rear in contact",
    "Contact", "Group2", "Group3", "Oral-oral Contact", "Oral-genital Contact", "Side by side Contact", "Side by side Contact, opposite way",
    "Train2", "FollowZone",
    "Social approach", "Approach contact",
    "Break contact"
    ]
    }

behaviouralEventOneMouseSingle = ["Move isolated", "Move in contact", "WallJump", "Stop isolated", "Rear isolated", "Rear in contact", "Oral-genital Contact", "Train2", "FollowZone",
                                "Social approach", "Approach contact", "Get away", "Break contact"]
behaviouralEventOneMouseSocial = ["Contact", "Group2", "Oral-oral Contact", "Oral-genital Contact",
                            "Side by side Contact", "Side by side Contact, opposite way",
                            "Train2", "FollowZone",
                            "Social approach", "Approach contact",
                            "Get away", "Break contact"]

categoryList = [' TotalLen', ' Nb', ' MeanDur']


class TestFileUtil ( unittest.TestCase ):
    
    def test_fillBetween(self):

        files = getFilesToProcess()
        for file in files:
            print ( file )
        
        self.assertEqual( True, True )